import typing

import pandas as pd
import polars as pl

from .._impl import TrieInnerNode
from ._NopTriePostprocessor import NopTriePostprocessor
from ._detail import TriePostprocessorBase


class CompoundTriePostprocessor(TriePostprocessorBase):
    """Functor to sequentially apply multiple trie postprocessors.

    Allows for the combination and sequential application of multiple trie
    postprocessors.
    """

    _postprocessors: typing.Iterable[typing.Callable]

    def __init__(
        self: "CompoundTriePostprocessor",
        postprocessors: typing.Iterable[TriePostprocessorBase],
    ) -> None:
        """Initialize functor instance.

        Parameters
        ----------
        postprocessors : typing.Iterable[TriePostprocessorBase]
            The sequence of postprocess functors to be applied.
        """
        self._postprocessors = postprocessors
        if not postprocessors:
            # ensure copy made if mutate is False
            self._postprocessors.append(NopTriePostprocessor())

    def __call__(
        self: "CompoundTriePostprocessor",
        trie: typing.Union[TrieInnerNode, pd.DataFrame, pl.DataFrame],
        p_differentia_collision: float,
        mutate: bool = False,
        progress_wrap: typing.Callable = lambda x: x,
    ) -> typing.Union[TrieInnerNode, pd.DataFrame, pl.DataFrame]:
        """Apply stored postprocessors in sequence.

        Parameters
        ----------
        trie : TrieInnerNode or pd.DataFrame or pl.DataFrame
            The input trie to be postprocessed.
        p_differentia_collision : float
            Probability of a randomly-generated differentia matching an
            existing differentia.

            Forwarded to the postprocess functors.
        mutate : bool, default False
            Are side effects on the input argument `trie` allowed?
        progress_wrap : typing.Callable, optional
            Pass tqdm or equivalent to report progress.

        Returns
        -------
        TrieInnerNode or pd.DataFrame or pl.DataFrame
            The postprocessed trie after applying all stored postprocessors.
        """
        for i, postprocessor in enumerate(self._postprocessors):
            trie = postprocessor(
                trie=trie,
                p_differentia_collision=p_differentia_collision,
                mutate=i > 0 or mutate,
                progress_wrap=progress_wrap,
            )

        return trie



import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_does_definitively_share_no_common_ancestor1(retention_policy, wrap):
    while True:
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
            stratum_retention_policy=retention_policy,
        )
        c2 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
            stratum_retention_policy=retention_policy,
        )
        if hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c1), wrap(c2)]
        ):
            assert hstrat.does_definitively_share_no_common_ancestor(
                [wrap(c2), wrap(c1)]
            )
            assert hstrat.does_definitively_share_no_common_ancestor(
                [wrap(c2), wrap(c1)] * 10
            )
            assert not hstrat.does_definitively_share_no_common_ancestor(
                [wrap(c1)] * 10
            )
            for __ in range(100):
                c1.DepositStratum()
                assert hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c1), wrap(c2)]
                )
                assert hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c2), wrap(c1)]
                )
                assert hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c2), wrap(c1)] * 10
                )
                assert hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c1)] * 10 + [wrap(c2)]
                )
            for __ in range(100):
                c2.DepositStratum()
                assert hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c1), wrap(c2)]
                )
                assert hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c2), wrap(c1)]
                )
                assert hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c2), wrap(c1)] * 10
                )
                assert hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c2)] * 10 + [wrap(c1)]
                )
            break

    while True:
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
            stratum_retention_policy=retention_policy,
        )
        c2 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=1,
            stratum_retention_policy=retention_policy,
        )
        if not hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c1), wrap(c2)]
        ):
            assert not hstrat.does_definitively_share_no_common_ancestor(
                [wrap(c2), wrap(c1)]
            )
            for __ in range(100):
                c1.DepositStratum()
                assert not hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c1), wrap(c2)]
                )
                assert not hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c2), wrap(c1)]
                )
                assert not hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c2), wrap(c1)] * 10
                )
                assert not hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c2)] * 10 + [wrap(c1)]
                )
            for __ in range(100):
                c2.DepositStratum()
                assert not hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c1), wrap(c2)]
                )
                assert not hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c2), wrap(c1)]
                )
                assert not hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c2), wrap(c1)] * 10
                )
                assert not hstrat.does_definitively_share_no_common_ancestor(
                    [wrap(c1)] * 10 + [wrap(c2)]
                )
            break

    for rep in range(100):
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=64,
            stratum_retention_policy=retention_policy,
        )
        c2 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=64,
            stratum_retention_policy=retention_policy,
        )
        assert hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c1), wrap(c2)]
        )
        assert hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c2), wrap(c1)]
        )
        assert hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c2), wrap(c1)] * 10
        )
        assert hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c1)] * 10 + [wrap(c2)]
        )


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_does_definitively_share_no_common_ancestor2(retention_policy, wrap):

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
        stratum_retention_policy=retention_policy,
    )
    c2 = c1.Clone()
    assert not hstrat.does_definitively_share_no_common_ancestor(
        [wrap(c1), wrap(c2)]
    )
    assert not hstrat.does_definitively_share_no_common_ancestor(
        [wrap(c2), wrap(c1)]
    )
    for __ in range(100):
        c1.DepositStratum()
        assert not hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c1), wrap(c2)]
        )
        assert not hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c2), wrap(c1)]
        )
    for __ in range(100):
        c2.DepositStratum()
        assert not hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c1), wrap(c2)]
        )
        assert not hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c2), wrap(c1)]
        )

    for rep in range(100):
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=64,
            stratum_retention_policy=retention_policy,
        )
        c2 = c1.Clone()
        assert not hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c1), wrap(c2)]
        )
        assert not hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c2), wrap(c1)]
        )


@pytest.mark.filterwarnings("ignore:Empty or singleton population.")
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_does_definitively_share_no_common_ancestor3(retention_policy, wrap):

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
        stratum_retention_policy=retention_policy,
    )
    assert (
        hstrat.does_definitively_share_no_common_ancestor([wrap(c1)]) is None
    )


@pytest.mark.filterwarnings("ignore:Empty or singleton population.")
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
def test_does_definitively_share_no_common_ancestor4(retention_policy):

    assert hstrat.does_definitively_share_no_common_ancestor([]) is None


@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_does_definitively_share_no_common_ancestor_generator(wrap):
    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=1,
    )
    for __ in range(10):
        assert hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c1) for __ in range(10)]
        ) == hstrat.does_definitively_share_no_common_ancestor(
            (wrap(c1) for __ in range(10))
        )
        c1.DepositStratum()

    c2 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
    )
    for __ in range(10):
        assert hstrat.does_definitively_share_no_common_ancestor(
            [wrap(c2) for __ in range(10)]
        ) == hstrat.does_definitively_share_no_common_ancestor(
            (wrap(c2) for __ in range(10))
        )
        c2.DepositStratum()



import typing

from ....._auxiliary_lib import popcount
from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec


class CalcNumStrataRetainedExact:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "CalcNumStrataRetainedExact",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "CalcNumStrataRetainedExact", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "CalcNumStrataRetainedExact",
        policy: PolicyCouplerBase,
        num_strata_deposited: int,
    ) -> int:
        """Exactly how many strata are retained after n deposited?

        The calculation can be written mathematically as,

          weight of binary expansion of n (i.e., #1's set in binary repr)
          + sum(
              floor( log2(n//r) )
              for r from 1 to r inclusive
          )
          + 1

        where

          n = num_strata_deposited - 1
          r = resolution

        This expression for exact number deposited was extrapolated from
            * resolution = 0, <https://oeis.org/A063787>
            * resolution = 1, <https://oeis.org/A056791>
        and is unit tested extensively.

        Note that the implementation must include a special case to account for
        n < r causing log2(0). In this case, the number of strata retained is
        equal to the number deposited (i.e., none have been discarded yet).
        """
        spec = policy.GetSpec()

        resolution = spec.GetRecencyProportionalResolution()
        if num_strata_deposited - 1 <= resolution:
            return num_strata_deposited
        else:
            return (
                # cast to int to handle numpy.int32, numpy.int64 etc.
                popcount(int(num_strata_deposited - 1))
                + sum(
                    # X.bit_length() - 1 equivalent to floor(log2(X))
                    # cast to int to handle numpy.int32, numpy.int64 etc.
                    int((num_strata_deposited - 1) // r).bit_length() - 1
                    for r in range(1, resolution + 1)
                )
                + 1
            )



from hstrat._auxiliary_lib import format_cli_description


def test_smoke():
    assert "asdf" in format_cli_description("asdf")
    assert "hello" in format_cli_description(
        """
hello.

world!
    1. one
    2. two
    3. three
"""
    )



import typing

from .._detail._PolicyCouplerBase import PolicyCouplerBase


class _GenDropRanksFromPredKeepRankBase:
    """Dummy class to faciliate recognition of instantiations of the
    GenDropRanksFromPredKeepRank class across different calls to
    the GenDropRanksFromPredKeepRank factory."""


def GenDropRanksFromPredKeepRank(
    predicate: typing.Type[typing.Callable],
) -> typing.Type[typing.Callable]:
    """Factory method to generate a stratum retention condemner functor from a
    stratum retention predicate functor.

    Parameters
    ----------
    predicate : callable
        Callable specifying whether a stratum with deposition rank r should be
        retained within the hereditary stratigraphic column after n strata have
        been deposited.

    Returns
    -------
    condemner : callable
        Functor that enacts the predicate's stratum retention policy by
        specifying the set of strata ranks that should be purged from a
        hereditary stratigraphic column when the nth stratum is deposited.
    """

    class GenDropRanksFromPredKeepRank(
        _GenDropRanksFromPredKeepRankBase,
    ):
        """Functor that wraps a stratum retention predicate functor to
        operate as a stratum retention condemner."""

        _predicate: typing.Callable

        def __init__(self: "GenDropRanksFromPredKeepRank", *args, **kwargs):
            """Construct the functor.

            Arguments are forwarded to predicate constructor.
            """
            self._predicate = predicate(*args, **kwargs)

        def __call__(
            self: "GenDropRanksFromPredKeepRank",
            policy: PolicyCouplerBase,
            num_stratum_depositions_completed: int,
            retained_ranks: typing.Iterable[int],
        ) -> typing.Iterator[int]:
            def should_retain(stratum_rank: int) -> bool:
                """Should the rth stratum rank be retained?

                Asserts retention requirements are respected by predicate.
                """
                res = self._predicate(
                    policy,
                    stratum_rank=stratum_rank,
                    num_stratum_depositions_completed=num_stratum_depositions_completed,
                )
                # predicate must *always* retain the initial and latest strata
                if stratum_rank in (0, num_stratum_depositions_completed):
                    assert res
                return res

            for rank in retained_ranks:
                if not should_retain(rank):
                    yield rank
            return

        def __eq__(
            self: "GenDropRanksFromPredKeepRank",
            other: typing.Any,
        ) -> bool:
            """Compare for value-wise equality."""
            # account for possible distinct instantiations of the
            # GenDropRanksFromPredKeepRank class across calls to the
            # GenDropRanksFromPredKeepRank factory
            if issubclass(
                other.__class__,
                _GenDropRanksFromPredKeepRankBase,
            ):
                return self._predicate == other._predicate
            else:
                return False

    return GenDropRanksFromPredKeepRank



from ._calc_common_ratio import calc_common_ratio
from ._get_retained_ranks import get_retained_ranks
from ._iter_rank_backstops import iter_rank_backstops
from ._iter_rank_cutoffs import iter_rank_cutoffs
from ._iter_rank_seps import iter_rank_seps
from ._iter_target_ranks import iter_target_ranks
from ._iter_target_recencies import iter_target_recencies

__all__ = [
    "calc_common_ratio",
    "get_retained_ranks",
    "iter_rank_backstops",
    "iter_rank_cutoffs",
    "iter_rank_seps",
    "iter_target_ranks",
    "iter_target_recencies",
]



import numpy as np
import pandas as pd


def alifestd_count_root_nodes(phylogeny_df: pd.DataFrame) -> np.ndarray:  # int
    """How many root nodes are contained in phylogeny?"""
    return (
        phylogeny_df["ancestor_list"]
        .astype(str)
        .str.lower()
        .isin(("[none]", "[]"))
        .sum()
    )



import numpy as np
import pytest

from hstrat import hstrat


def test_estimate_rank_of_mrca_unbiased_no_common_ancestor():
    coincident_ranks = iter([0])
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_unbiased(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    assert result is None


def test_estimate_rank_of_mrca_unbiased_with_common_ancestor1():
    coincident_ranks = iter([1, 0])
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_unbiased(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    assert result == 0


def test_estimate_rank_of_mrca_unbiased_with_common_ancestor2():
    coincident_ranks = iter([7, 3, 0])
    p_differentia_collision = 0.5

    result = hstrat.estimate_rank_of_mrca_unbiased(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    expected_result = np.average([4.5, 1], weights=[2, 1])
    assert np.isclose(result, expected_result)


def test_estimate_rank_of_mrca_unbiased_with_common_ancestor2b():
    coincident_ranks = iter([7, 3, 0])
    p_differentia_collision = 0.25

    result = hstrat.estimate_rank_of_mrca_unbiased(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    expected_result = np.average([4.5, 1], weights=[4, 1])
    assert np.isclose(result, expected_result)


def test_estimate_rank_of_mrca_unbiased_with_common_ancestor3():
    coincident_ranks = [8, 7, 2, 0]
    p_differentia_collision = 0.5

    expected_result = np.average([7, 4, 0.5], weights=[4, 2, 1])
    result = hstrat.estimate_rank_of_mrca_unbiased(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    assert np.isclose(result, expected_result)


@pytest.mark.parametrize(
    "prior",
    [
        hstrat.ArbitraryPrior(),
        hstrat.UniformPrior(),
        hstrat.ExponentialPrior(1.1),
        hstrat.GeometricPrior(1.1),
    ],
)
@pytest.mark.parametrize(
    "p_differentia_collision", [0.5, 0.25, 0.5**8, 0.5**64]
)
def test_estimate_rank_of_mrca_unbiased_with_common_ancestor4(
    prior, p_differentia_collision
):
    coincident_ranks = iter([8, 7, 2, 0])

    # Calculate expected result using np.average with weights
    expected_ranks = [
        prior.CalcIntervalConditionedMean(7, 8),
        prior.CalcIntervalConditionedMean(2, 7),
        prior.CalcIntervalConditionedMean(0, 2),
    ]
    weights = [
        p_differentia_collision**0
        * prior.CalcIntervalProbabilityProxy(7, 8),
        p_differentia_collision**1
        * prior.CalcIntervalProbabilityProxy(2, 7),
        p_differentia_collision**2
        * prior.CalcIntervalProbabilityProxy(0, 2),
    ]
    expected_result = np.average(expected_ranks, weights=weights)

    result = hstrat.estimate_rank_of_mrca_unbiased(
        coincident_ranks, p_differentia_collision, prior
    )

    assert np.isclose(result, expected_result)


def test_estimate_rank_of_mrca_unbiased_with_common_ancestor5():
    coincident_ranks = [8, 7, 2, 0]
    p_differentia_collision = 0.5**64

    result = hstrat.estimate_rank_of_mrca_unbiased(
        coincident_ranks, p_differentia_collision, hstrat.ArbitraryPrior()
    )

    assert np.isclose(result, 7)


def test_estimate_rank_of_mrca_unbiased_with_common_ancestor6():
    coincident_ranks = [7, 3, 0]
    p_differentia_collision = 0.5

    result1 = hstrat.estimate_rank_of_mrca_unbiased(
        coincident_ranks, p_differentia_collision, hstrat.GeometricPrior(2)
    )
    result2 = hstrat.estimate_rank_of_mrca_unbiased(
        coincident_ranks, p_differentia_collision, hstrat.GeometricPrior(2.1)
    )

    assert 4 < result1 < result2 < 6



from .._detail import PolicyCouplerFactory
from ._PolicySpec import PolicySpec
from ._enact._GenDropRanks import GenDropRanks
from ._invar._CalcMrcaUncertaintyAbsUpperBound import (
    CalcMrcaUncertaintyAbsUpperBound,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyAbsUpperBoundPessimalRank import (
    CalcMrcaUncertaintyAbsUpperBoundPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBound import (
    CalcMrcaUncertaintyRelUpperBound,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundAtPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
)
from ._invar._CalcMrcaUncertaintyRelUpperBoundPessimalRank import (
    CalcMrcaUncertaintyRelUpperBoundPessimalRank,
)
from ._invar._CalcNumStrataRetainedUpperBound import (
    CalcNumStrataRetainedUpperBound,
)
from ._scry._CalcMrcaUncertaintyAbsExact import CalcMrcaUncertaintyAbsExact
from ._scry._CalcMrcaUncertaintyRelExact import CalcMrcaUncertaintyRelExact
from ._scry._CalcNumStrataRetainedExact import CalcNumStrataRetainedExact
from ._scry._CalcRankAtColumnIndex import CalcRankAtColumnIndex
from ._scry._IterRetainedRanks import IterRetainedRanks

Policy = PolicyCouplerFactory(
    policy_spec_t=PolicySpec,
    # enactment
    gen_drop_ranks_ftor_t=GenDropRanks,
    # invariants
    calc_mrca_uncertainty_abs_upper_bound_ftor_t=CalcMrcaUncertaintyAbsUpperBound,
    calc_mrca_uncertainty_abs_upper_bound_at_pessimal_rank_ftor_t=CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank,
    calc_mrca_uncertainty_abs_upper_bound_pessimal_rank_ftor_t=CalcMrcaUncertaintyAbsUpperBoundPessimalRank,
    calc_mrca_uncertainty_rel_upper_bound_ftor_t=CalcMrcaUncertaintyRelUpperBound,
    calc_mrca_uncertainty_rel_upper_bound_at_pessimal_rank_ftor_t=CalcMrcaUncertaintyRelUpperBoundAtPessimalRank,
    calc_mrca_uncertainty_rel_upper_bound_pessimal_rank_ftor_t=CalcMrcaUncertaintyRelUpperBoundPessimalRank,
    calc_num_strata_retained_upper_bound_ftor_t=CalcNumStrataRetainedUpperBound,
    # scrying
    calc_mrca_uncertainty_abs_exact_ftor_t=CalcMrcaUncertaintyAbsExact,
    calc_mrca_uncertainty_rel_exact_ftor_t=CalcMrcaUncertaintyRelExact,
    calc_num_strata_retained_exact_ftor_t=CalcNumStrataRetainedExact,
    calc_rank_at_column_index_ftor_t=CalcRankAtColumnIndex,
    iter_retained_ranks_ftor_t=IterRetainedRanks,
)

# gloss away PolicyCoupler implementation details
Policy.__name__ = "Policy"
Policy.__qualname__ = "Policy"
Policy.__module__ = __name__



import numpy as np
import pandas as pd
import pytest

from hstrat._auxiliary_lib import get_nullable_mask


@pytest.mark.parametrize("dtype_base", ["UInt{}", "Int{}"])
@pytest.mark.parametrize("dtype_bits", [8, 16, 32, 64])
def test_get_nullable_mask(dtype_base, dtype_bits):
    series = pd.Series([3, 2, pd.NA], dtype=dtype_base.format(dtype_bits))
    assert np.array_equal(get_nullable_mask(series), series.isna())
    get_nullable_mask(series)[0] = True
    assert np.array_equal(series.isna(), np.array([True, False, True]))
    get_nullable_mask(series)[2] = False
    assert np.array_equal(series.isna(), np.array([True, False, False]))



import numpy as np
import pandas as pd

from hstrat._auxiliary_lib import (
    alifestd_reroot_at_id_asexual,
    alifestd_validate,
)


def test_alifestd_reroot_at_id_asexual_inner_node_terminal_node():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [3, 3, 4, 4, 5, 5],
            "ancestor_list": ["[3]", "[3]", "[4]", "[4]", "[5]", "[none]"],
        }
    )
    assert alifestd_validate(phylogeny_df)
    reroot_at = 1
    expected_result = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [3, 1, 4, 1, 3, 4],
            "ancestor_list": ["[3]", "[none]", "[4]", "[1]", "[3]", "[4]"],
        }
    )

    for df in (phylogeny_df, phylogeny_df.sample(frac=1)):
        df_ = df.copy()
        result = alifestd_reroot_at_id_asexual(df, reroot_at)

        assert expected_result.equals(result.sort_index())
        assert df.equals(df_)


def test_alifestd_reroot_at_id_asexual_inner_node_bifurcation():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [1, 3, 4, 4, 4, 1],
            "ancestor_list": ["[1]", "[3]", "[4]", "[4]", "[none]", "[1]"],
        }
    )
    assert alifestd_validate(phylogeny_df)
    reroot_at = 1
    expected_result = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [1, 1, 4, 1, 3, 1],
            "ancestor_list": ["[1]", "[none]", "[4]", "[1]", "[3]", "[1]"],
        }
    )

    for df in (phylogeny_df, phylogeny_df.sample(frac=1)):
        df_ = df.copy()
        result = alifestd_reroot_at_id_asexual(df, reroot_at)

        assert expected_result.equals(result.sort_index())
        assert df.equals(df_)


def test_alifestd_reroot_at_id_asexual_inner_node_trifurcation():
    phylogeny_df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [1, 3, 1, 4, 4, 1],
            "ancestor_list": ["[1]", "[3]", "[1]", "[4]", "[none]", "[1]"],
        }
    )
    assert alifestd_validate(phylogeny_df)
    reroot_at = 1
    expected_result = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [1, 1, 1, 1, 3, 1],
            "ancestor_list": ["[1]", "[none]", "[1]", "[1]", "[3]", "[1]"],
        }
    )

    for df in (phylogeny_df, phylogeny_df.sample(frac=1)):
        df_ = df.copy()
        result = alifestd_reroot_at_id_asexual(df, reroot_at)

        assert expected_result.equals(result.sort_index())
        assert df.equals(df_)


def test_alifestd_reroot_at_id_asexual_singleton():
    phylogeny_df = pd.DataFrame(
        {
            "id": [7],
            "ancestor_id": [7],
            "ancestor_list": "[none]",
        }
    )
    reroot_at = 7
    expected_result = phylogeny_df

    df_ = phylogeny_df.copy()
    result = alifestd_reroot_at_id_asexual(phylogeny_df, reroot_at)
    assert np.array_equal(expected_result, result)
    assert phylogeny_df.equals(df_)



from matplotlib import pyplot as plt


# adapted from https://stackoverflow.com/a/61039272
# fixes warning
# Figures created through the pyplot interface (`matplotlib.pyplot.figure`) are
# retained until explicitly closed and may consume too much memory.
def release_cur_mpl_fig():
    plt.clf()
    plt.close()



import logging
import os
import sys


def configure_prod_logging() -> None:
    logging.basicConfig(
        datefmt="%Y-%m-%d %H:%M:%S",
        format="%(asctime)s %(levelname)-8s %(message)s",
        level=logging.INFO,
    )
    if not sys.__stdout__.isatty():
        os.environ["TQDM_MININTERVAL"] = "5"



"""Maintain constant size complexity and evenly-spaced strata.

The tapered depth-proportional resolution policy ensures estimates of MRCA
rank will have uncertainty bounds less than or equal to a user-specified
proportion of the largest number of strata deposited on either column. Thus,
MRCA rank estimate uncertainty scales as O(n) with respect to the greater
number of strata deposited on either column.

Under the tapered depth-proportional resolution policy, the number of strata
retained (i.e., space complexity) scales as O(1) with respect to the number of
strata deposited.

This functor's retention policy implementation guarantees that columns will
retain appropriate strata so that for any two columns with m and n strata
deposited, the rank of the most recent common ancestor can be determined with
uncertainty of at most

    bound = floor(max(m, n) / depth_proportional_resolution)

ranks. Achieving this limit on uncertainty requires retaining sufficient strata
so that no more than bound ranks elapsed between any two strata. This policy
accumulates retained strata at a fixed interval until twice as many as
depth_proportional_resolution are at hand. Then, every other
retained stratum is purged gradually from back to front until the cycle repeats
with a new twice-as-wide interval between retained strata.

See Also
--------
StratumRetentionPredicateDepthProportionalResolution:
    For a predicate retention policy that achieves the same guarantees for
    tapered depth-proportional resolution but purges unnecessary strata more
    aggressively and abruptly.
"""

from ...._auxiliary_lib import lazy_attach_stub

__getattr__, __dir__, __all__ = lazy_attach_stub(__name__, __file__)
del lazy_attach_stub



import itertools as it
import random

from iterify import cyclify, iterify
import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.95, 0.88],
)
def test_calc_patristic_distance_bounds_between_specimen(
    retention_policy, differentia_width, confidence_level
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.calc_patristic_distance_bounds_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column2),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_patristic_distance_bounds_between(
        column, column2, prior="arbitrary", confidence_level=confidence_level
    )
    assert hstrat.calc_patristic_distance_bounds_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_patristic_distance_bounds_between(
        column, column, prior="arbitrary", confidence_level=confidence_level
    )

    assert hstrat.calc_patristic_distance_bounds_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(child1),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_patristic_distance_bounds_between(
        column, child1, prior="arbitrary", confidence_level=confidence_level
    )

    assert hstrat.calc_patristic_distance_bounds_between(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_patristic_distance_bounds_between(
        child1, child2, prior="arbitrary", confidence_level=confidence_level
    )

    child1.DepositStrata(10)
    assert hstrat.calc_patristic_distance_bounds_between(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        prior="arbitrary",
        confidence_level=confidence_level,
    ) == hstrat.calc_patristic_distance_bounds_between(
        child1, child2, prior="arbitrary", confidence_level=confidence_level
    )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_CalcRanksSinceMrcaBoundsWith(
    retention_policy,
):
    def make_bundle():
        return hstrat.HereditaryStratigraphicColumnBundle(
            {
                "test": hstrat.HereditaryStratigraphicColumn(
                    initial_stratum_annotation=0,
                    stratum_retention_policy=retention_policy,
                ),
                "control": hstrat.HereditaryStratigraphicColumn(
                    initial_stratum_annotation=0,
                    stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
                ),
            }
        )

    column = make_bundle()
    frozen_copy = column.Clone()
    frozen_unrelated = make_bundle()
    population = [column.Clone() for __ in range(10)]
    forked_isolated = column.Clone()
    unrelated_isolated = make_bundle()

    for _generation in range(100):
        _ = _generation
        for f, s in it.chain(
            it.combinations(population, 2),
            zip(population, cyclify(forked_isolated)),
            zip(population, cyclify(frozen_copy)),
            zip(cyclify(forked_isolated), population),
            zip(cyclify(frozen_copy), population),
        ):
            lb, ub = hstrat.calc_patristic_distance_bounds_between(
                f["test"],
                s["test"],
                prior="arbitrary",
            )
            actual_rank_of_mrca = hstrat.get_last_common_stratum_between(
                f["control"],
                s["control"],
            ).GetAnnotation()
            actual_patristic_distance = (
                f.GetNumStrataDeposited() - actual_rank_of_mrca - 1
            ) + (s.GetNumStrataDeposited() - actual_rank_of_mrca - 1)
            assert lb <= actual_patristic_distance < ub

        for f, s in it.chain(
            zip(population, cyclify(frozen_unrelated)),
            zip(population, cyclify(unrelated_isolated)),
            zip(cyclify(frozen_unrelated), population),
            zip(cyclify(unrelated_isolated), population),
        ):
            assert (
                hstrat.calc_patristic_distance_bounds_between(
                    f["test"],
                    s["test"],
                    prior="arbitrary",
                )
                is None
            )

        # advance generation
        random.shuffle(population)
        for target in range(3):
            population[target] = population[-1].Clone()
            population[target].DepositStratum(
                annotation=population[target].GetNumStrataDeposited(),
            )
        for individual in it.chain(
            iter(population),
            iterify(forked_isolated),
            iterify(unrelated_isolated),
        ):
            if random.choice([True, False]):
                individual.DepositStratum(
                    annotation=individual.GetNumStrataDeposited(),
                )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_comparison_commutativity_asynchronous(
    differentia_width,
    retention_policy,
):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(30):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.calc_patristic_distance_bounds_between(
                first,
                second,
                prior="arbitrary",
            ) == hstrat.calc_patristic_distance_bounds_between(
                second,
                first,
                prior="arbitrary",
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].CloneDescendant()
        for individual in population:
            # asynchronous generations
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy_2a,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_comparison_commutativity_synchronous(
    differentia_width,
    retention_policy,
):

    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.combinations(population, 2):
            # assert commutativity
            assert hstrat.calc_patristic_distance_bounds_between(
                first,
                second,
                prior="arbitrary",
            ) == hstrat.calc_patristic_distance_bounds_between(
                second,
                first,
                prior="arbitrary",
            )

        # advance generation
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        # synchronous generations
        for individual in population:
            individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "retention_policy",
    [
        pytest.param(
            hstrat.perfect_resolution_algo.Policy(),
            marks=pytest.mark.heavy,
        ),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=10,
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
def test_comparison_validity(retention_policy, differentia_width):
    population = [
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_width,
            stratum_retention_policy=retention_policy,
        )
        for __ in range(10)
    ]

    for _generation in range(100):
        _ = _generation
        for first, second in it.permutations(population, 2):
            rsmw1 = hstrat.calc_ranks_since_mrca_bounds_with(
                first,
                second,
                prior="arbitrary",
            )
            rsmw2 = hstrat.calc_ranks_since_mrca_bounds_with(
                second,
                first,
                prior="arbitrary",
            )
            pd = hstrat.calc_patristic_distance_bounds_between(
                first,
                second,
                prior="arbitrary",
            )
            assert (rsmw1 is None) == (rsmw1 is None) == (pd is None)
            if pd is not None:
                assert all(
                    pytest.approx(pd_bound) == rsmw1_bound + rsmw2_bound
                    for rsmw1_bound, rsmw2_bound, pd_bound in zip(
                        rsmw1, rsmw2, pd
                    )
                )

        # advance generations asynchronously
        random.shuffle(population)
        for target in range(5):
            population[target] = population[-1].Clone()
        for individual in population:
            if random.choice([True, False]):
                individual.DepositStratum()


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary"],
)
def test_artifact_types_equiv(differentia_width, policy, prior):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.calc_patristic_distance_bounds_between(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
            prior,
        ) == hstrat.calc_patristic_distance_bounds_between(a, b, prior)



import numbers
import typing

from ._intersect_ranges import intersect_ranges


def assign_intersecting_subsets(
    ranges: typing.List[typing.Tuple[numbers.Real, numbers.Real]],
) -> typing.List[typing.List[numbers.Integral]]:
    """Arrange ranges into groups such that all ranges within a group share at
    least one overlapping value.

    Composes groups such that intersected values are as low as possible. Put
    another way, ranges are preferentially associated with earlier groups.

    Returns groups with constituent ranges' indices in `ranges` instead of the
    ranges themselves.

    Parameters
    ----------
    ranges : list of tuples
        A list of tuples, each containing two real numbers representing the
        lwoer bound (inclusive) and upper bound (exclusive) of a range.

    Returns
    -------
    list of lists of tuples
        A list of lists, each containing integer values representing the
        indices of intersecting ranges within a group.


    Examples
    --------
    >>> make_intersecting_subsets([(1, 3), (5, 6), (2, 4)])
    [[0, 2], [1]]
    """

    range_subsets = []
    index_subsets = []
    for range_index, range_ in sorted(enumerate(ranges), key=lambda x: x[1]):
        for range_subset, index_subset in zip(range_subsets, index_subsets):
            if intersect_ranges([range_, *range_subset]):
                range_subset.append(range_)
                index_subset.append(range_index)
                break
        else:
            range_subsets.append([range_])
            index_subsets.append([range_index])

    return index_subsets



import numpy as np
import pandas as pd

from ._alifestd_parse_ancestor_ids import alifestd_parse_ancestor_ids


def alifestd_has_increasing_ids(phylogeny_df: pd.DataFrame) -> bool:
    """Do offspring have larger id values than ancestors?

    Input dataframe is not mutated by this operation.
    """
    if "ancestor_id" in phylogeny_df:
        return np.all(phylogeny_df["id"] >= phylogeny_df["ancestor_id"])
    else:
        for _idx, row in phylogeny_df.iterrows():
            ancestor_list = alifestd_parse_ancestor_ids(row["ancestor_list"])
            if len(ancestor_list) == 0:
                continue
            elif max(ancestor_list) > row["id"]:
                return False

        return True



import typing

import pandas as pd
import sortedcontainers as sc

from ._alifestd_has_contiguous_ids import alifestd_has_contiguous_ids
from ._alifestd_is_topologically_sorted import alifestd_is_topologically_sorted
from ._alifestd_mark_num_descendants_asexual import (
    alifestd_mark_num_descendants_asexual,
)
from ._alifestd_topological_sort import alifestd_topological_sort
from ._alifestd_try_add_ancestor_id_col import alifestd_try_add_ancestor_id_col


def alifestd_categorize_triplet_asexual(
    phylogeny_df: pd.DataFrame,
    triplet_ids: typing.Iterable[int],
    mutate: bool = False,
) -> int:
    """Assess the topological configuration of three `id`'s in `phylogeny_df`.

    If polytomy, return -1. Else, return index of outgroup `id`.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.

    See Also
    --------
    alifestd_estimate_triplet_distance_asexual
    alifestd_sample_triplet_comparisons_asexual
    """

    if not mutate:
        phylogeny_df = phylogeny_df.copy()

    phylogeny_df = alifestd_try_add_ancestor_id_col(phylogeny_df, mutate=True)

    if not alifestd_is_topologically_sorted(phylogeny_df):
        phylogeny_df = alifestd_topological_sort(phylogeny_df, mutate=True)

    if "num_descendants" not in phylogeny_df:
        phylogeny_df = alifestd_mark_num_descendants_asexual(
            phylogeny_df, mutate=True
        )

    if alifestd_has_contiguous_ids(phylogeny_df):
        phylogeny_df.reset_index(drop=True, inplace=True)
    else:
        phylogeny_df.index = phylogeny_df["id"]

    trace = [*triplet_ids]
    queue = sc.SortedSet(trace, key=phylogeny_df.index.get_loc)
    if not len(queue) == 3:
        raise ValueError(
            "triplet_ids should have 3 unique values, "
            f"but only has {len(queue)}.",
        )

    while len(queue) == 3 or trace.count(queue[-1]) == 1:
        assert len(queue) >= 2
        oldest = queue.pop(-1)
        replacement = phylogeny_df.at[oldest, "ancestor_id"]
        assert replacement != oldest
        queue.add(replacement)
        trace[trace.index(oldest)] = replacement

    counts = [*map(trace.count, trace)]
    assert counts.count(1) <= 1
    try:
        return counts.index(1)
    except ValueError:
        return -1



import pandas as pd
import pytest

from hstrat._auxiliary_lib import (
    alifestd_calc_polytomic_index,
    alifestd_make_empty,
)


def test_empty_df():
    assert alifestd_calc_polytomic_index(alifestd_make_empty()) == 0


def test_singleton_df():
    df = pd.DataFrame(
        {
            "id": [0],
            "ancestor_list": [[None]],
        }
    )
    assert alifestd_calc_polytomic_index(df) == 0


def test_polytomy_df1():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4],
            "ancestor_list": [[None], [0], [0], [0], [1]],
        }
    )
    assert alifestd_calc_polytomic_index(df) == 1


def test_polytomy_df2():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [[None], [0], [0], [0], [1], [1], [1]],
        }
    )
    assert alifestd_calc_polytomic_index(df) == 2


def test_polytomy_df3():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7],
            "ancestor_list": [[None], [0], [0], [0], [1], [1], [1], [1]],
        }
    )
    assert alifestd_calc_polytomic_index(df) == 3


def test_multiple_trees_df1():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": [[None], [None], [0], [2], [2], [3]],
        }
    )
    assert alifestd_calc_polytomic_index(df) == 0


def test_multiple_trees_df2():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_list": [[None], [None], [0], [1], [0], [0]],
        }
    )
    assert alifestd_calc_polytomic_index(df) == 1


def test_multiple_trees_df3():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_list": [[None], [None], [0], [1], [0], [0], [None]],
        }
    )
    assert alifestd_calc_polytomic_index(df) == 1


def test_sexual():
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 5],
            "ancestor_list": [[None], [None], [0, 1], [1], [0]],
        }
    )
    with pytest.raises(ValueError):
        alifestd_calc_polytomic_index(df)



"""Functions to reconstruct a phylogenetic tree from extant hereditary strata."""

from . import trie_postprocess
from ..._auxiliary_lib import lazy_attach

__getattr__, __dir__, __all__ = lazy_attach(
    __name__,
    submodules=["trie_postprocess"],
    submod_attrs={
        "trie_postprocess": trie_postprocess.__all__,
        "_build_tree": ["build_tree"],
        "_build_tree_nj": ["build_tree_nj"],
        "_build_tree_searchtable": ["build_tree_searchtable"],
        "_build_tree_trie": ["build_tree_trie"],
        "_build_tree_trie_ensemble": ["build_tree_trie_ensemble"],
        "_build_tree_upgma": ["build_tree_upgma"],
    },
    should_launder=[
        "build_tree_nj",
        "build_tree_searchtable",
        "build_tree_trie",
        "build_tree_trie_ensemble",
        "build_tree_upgma",
        "build_tree",
        "trie_postprocess",
    ].__contains__,
)
del lazy_attach



import typing

from ...genome_instrumentation import HereditaryStratigraphicColumn
from ._iter_mutual_ranks import iter_mutual_ranks


def iter_ranks_of_retained_commonality_between(
    first: HereditaryStratigraphicColumn,
    second: HereditaryStratigraphicColumn,
    *,
    first_start_idx: int = 0,
    second_start_idx: int = 0,
) -> typing.Iterator[int]:
    """Iterate over ranks with matching strata between columns in ascending
    order."""
    for rank, comp in iter_mutual_ranks(
        first,
        second,
        compare=True,
        first_start_idx=first_start_idx,
        second_start_idx=second_start_idx,
    ):
        if comp:
            yield rank
        else:
            return



import typing

from ..._detail import PolicyCouplerBase
from .._PolicySpec import PolicySpec


class IterRetainedRanks:
    """Functor to provide member function implementation in Policy class."""

    def __init__(
        self: "IterRetainedRanks",
        policy_spec: typing.Optional[PolicySpec],
    ) -> None:
        pass

    def __eq__(self: "IterRetainedRanks", other: typing.Any) -> bool:
        return isinstance(other, self.__class__)

    def __call__(
        self: "IterRetainedRanks",
        policy: PolicyCouplerBase,
        num_strata_deposited: int,
    ) -> typing.Iterator[int]:
        """Iterate over retained strata ranks at `num_strata_deposited` in
        ascending order."""
        spec = policy.GetSpec()

        yield from range(0, num_strata_deposited, spec.GetFixedResolution())

        last_rank = num_strata_deposited - 1
        if last_rank > 0 and last_rank % spec.GetFixedResolution():
            yield last_rank



import pandas as pd

from hstrat._auxiliary_lib import (
    alifestd_categorize_triplet_asexual,
    alifestd_make_ancestor_list_col,
)


def test_polytomy():
    df = pd.DataFrame(
        {
            "id": [1, 2, 3],
            "ancestor_id": [1, 1, 1],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df, df.sample(frac=1), df.sample(frac=1), df.sample(frac=1):
        assert alifestd_categorize_triplet_asexual(xdf, [1, 2, 3]) == -1
        assert alifestd_categorize_triplet_asexual(xdf, [3, 1, 2]) == -1

    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 0, 0],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df, df.sample(frac=1), df.sample(frac=1), df.sample(frac=1):
        assert alifestd_categorize_triplet_asexual(xdf, [1, 2, 3]) == -1
        assert alifestd_categorize_triplet_asexual(xdf, [3, 1, 2]) == -1

    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [0, 0, 0, 0, 1, 2],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df, df.sample(frac=1), df.sample(frac=1), df.sample(frac=1):
        assert alifestd_categorize_triplet_asexual(xdf, [3, 4, 5]) == -1
        assert alifestd_categorize_triplet_asexual(xdf, [5, 3, 4]) == -1
        assert alifestd_categorize_triplet_asexual(xdf, [5, 4, 3]) == -1

    df = pd.DataFrame(
        {
            "id": [9, 1, 2, 3, 4, 5],
            "ancestor_id": [9, 9, 9, 9, 1, 2],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df, df.sample(frac=1), df.sample(frac=1), df.sample(frac=1):
        assert alifestd_categorize_triplet_asexual(xdf, [3, 4, 5]) == -1
        assert alifestd_categorize_triplet_asexual(xdf, [5, 3, 4]) == -1
        assert alifestd_categorize_triplet_asexual(xdf, [5, 4, 3]) == -1

    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5, 6],
            "ancestor_id": [0, 0, 0, 0, 1, 2, 3],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df, df.sample(frac=1), df.sample(frac=1), df.sample(frac=1):
        assert alifestd_categorize_triplet_asexual(xdf, [6, 4, 5]) == -1
        assert alifestd_categorize_triplet_asexual(xdf, [5, 6, 4]) == -1
        assert alifestd_categorize_triplet_asexual(xdf, [5, 4, 6]) == -1


def test_bifurcating():
    df = pd.DataFrame(
        {
            "id": [1, 2, 3, 4, 5],
            "ancestor_id": [1, 1, 2, 2, 1],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df, df.sample(frac=1), df.sample(frac=1), df.sample(frac=1):
        assert alifestd_categorize_triplet_asexual(xdf, [3, 4, 5]) == 2
        assert alifestd_categorize_triplet_asexual(xdf, [3, 5, 4]) == 1
        assert alifestd_categorize_triplet_asexual(xdf, [5, 3, 4]) == 0

    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "ancestor_id": [0, 0, 1, 2, 2, 1],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df, df.sample(frac=1), df.sample(frac=1), df.sample(frac=1):
        assert alifestd_categorize_triplet_asexual(xdf, [3, 4, 5]) == 2
        assert alifestd_categorize_triplet_asexual(xdf, [3, 5, 4]) == 1
        assert alifestd_categorize_triplet_asexual(xdf, [5, 3, 4]) == 0

    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 6, 5, 4],
            "ancestor_id": [0, 0, 1, 2, 2, 1, 6],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df, df.sample(frac=1), df.sample(frac=1), df.sample(frac=1):
        assert alifestd_categorize_triplet_asexual(xdf, [3, 4, 5]) == 2
        assert alifestd_categorize_triplet_asexual(xdf, [3, 5, 4]) == 1
        assert alifestd_categorize_triplet_asexual(xdf, [5, 3, 4]) == 0

    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3, 4, 5],
            "taxon_label": ["0", "1", "2", "3", "4", "5"],
            "ancestor_id": [0, 0, 1, 2, 1, 2],
        },
    )
    df["ancestor_list"] = alifestd_make_ancestor_list_col(
        df["id"], df["ancestor_id"]
    )
    for xdf in df, df.sample(frac=1), df.sample(frac=1), df.sample(frac=1):
        assert alifestd_categorize_triplet_asexual(xdf, [3, 4, 5]) == 1
        assert alifestd_categorize_triplet_asexual(xdf, [3, 5, 4]) == 2
        assert alifestd_categorize_triplet_asexual(xdf, [4, 3, 5]) == 0



import numpy as np
import typing_extensions


def bytes_swap_bit_order(buffer: typing_extensions.Buffer) -> bytes:
    r"""Reverse the bit order of each byte in a given buffer.

    Parameters
    ----------
    buffer : typing_extensions.Buffer
        Input buffer whose bits in each byte are to be reversed.

    Returns
    -------
    bytes
        A new buffer with the bit order of each byte in the input buffer
        reversed.

    Examples
    --------
    >>> input_bytes = b'\x01\x02'
    >>> bytes_swap_bit_order(input_bytes)
    b'\x80\x40'
    """
    numpy_bytes = np.frombuffer(buffer, dtype=np.uint8)
    numpy_bits = np.unpackbits(numpy_bytes, bitorder="little")
    return np.packbits(numpy_bits, bitorder="big").tobytes()



import itertools as it

import numpy as np
import pytest

from hstrat.hstrat import recency_proportional_resolution_algo as rpra


@pytest.mark.parametrize(
    "recency_proportional_resolution",
    [
        0,
        1,
        2,
        3,
        7,
        42,
        97,
        100,
    ],
)
@pytest.mark.parametrize(
    "num_stratum_depositions_completed",
    it.chain(
        map(
            int,
            np.random.default_rng(1).integers(
                2**63,
                size=20,
            ),
        ),
        range(100),
        range(0, 10000, 1007),
    ),
)
def test_impl_consistency(
    recency_proportional_resolution: int,
    num_stratum_depositions_completed: int,
):
    assert len(
        {
            impl(
                recency_proportional_resolution,
                num_stratum_depositions_completed,
            )
            for impl in rpra._impl._num_to_condemn_.impls
        }
    )



import typing

import pandas as pd
import polars as pl
import pytest

from hstrat import hstrat
from hstrat._auxiliary_lib import coerce_to_pandas
import hstrat.phylogenetic_inference.tree.trie_postprocess._detail as detail


def test_base_class():
    assert issubclass(
        hstrat.AssignOriginTimeNodeRankTriePostprocessor,
        detail.TriePostprocessorBase,
    )


@pytest.mark.parametrize("df_type", [pd.DataFrame, pl.DataFrame])
def test_assign_trie_origin_times_node_rank_single_leaf(df_type: typing.Type):
    df = df_type(
        {
            "id": [0, 1, 2],
            "ancestor_id": [0, 0, 1],
            "rank": [0, 0, 0],
            "taxon_label": [None, None, "A"],
            "differentia": [None, 0, None],
        },
    )
    df = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        df, p_differentia_collision=0.5, mutate=True
    )
    df = coerce_to_pandas(df)
    assert df.loc[0, "origin_time"] == 0
    assert df.loc[1, "origin_time"] == 0
    assert df.loc[2, "origin_time"] == 0


@pytest.mark.parametrize("df_type", [pd.DataFrame, pl.DataFrame])
def test_assign_trie_origin_times_node_rank_two_leaves(df_type: typing.Type):
    df = df_type(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 2],
            "rank": [0, 0, 0, 0],
            "taxon_label": [None, None, "A", "B"],
            "differentia": [None, 0, None, None],
            "t0_": [1, 1, 1, 1],
        },
    )
    df = hstrat.AssignOriginTimeNodeRankTriePostprocessor(t0="t0_")(
        df, p_differentia_collision=0.5, mutate=True
    )
    df = coerce_to_pandas(df)
    assert df.loc[0, "origin_time"] == 0 - 1
    assert df.loc[1, "origin_time"] == 0 - 1
    assert df.loc[2, "origin_time"] == 0 - 1
    assert df.loc[3, "origin_time"] == 0 - 1


@pytest.mark.parametrize("df_type", [pd.DataFrame, pl.DataFrame])
def test_assign_trie_origin_times_node_rank1(df_type: typing.Type):
    df = df_type(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 2],
            "rank": [0, 0, 1, 1],
            "taxon_label": [None, None, "A", "B"],
            "differentia": [None, 0, None, None],
        },
    )
    df = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        df, p_differentia_collision=0.5, mutate=True
    )
    df = coerce_to_pandas(df)
    assert df.loc[0, "origin_time"] == 0
    assert df.loc[1, "origin_time"] == 0
    assert df.loc[2, "origin_time"] == 1
    assert df.loc[3, "origin_time"] == 1


@pytest.mark.parametrize("df_type", [pd.DataFrame, pl.DataFrame])
def test_assign_trie_origin_times_node_rank2(df_type: typing.Type):
    df = df_type(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 2],
            "rank": [0, 0, 2, 2],
            "taxon_label": [None, None, "A", "B"],
            "differentia": [None, 0, None, None],
        },
    )
    df = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        df, p_differentia_collision=0.5, mutate=True
    )
    df = coerce_to_pandas(df)
    assert df.loc[0, "origin_time"] == 0
    assert df.loc[1, "origin_time"] == 0
    assert df.loc[2, "origin_time"] == 2
    assert df.loc[3, "origin_time"] == 2


@pytest.mark.parametrize("df_type", [pd.DataFrame, pl.DataFrame])
def test_assign_trie_origin_times_node_rank3(df_type: typing.Type):
    df = df_type(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 2],
            "rank": [0, 0, 3, 3],
            "taxon_label": [None, None, "A", "B"],
            "differentia": [None, 0, 10, None],
        },
    )
    df = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        df, p_differentia_collision=0.5, mutate=True
    )
    df = coerce_to_pandas(df)
    assert df.loc[0, "origin_time"] == 0
    assert df.loc[1, "origin_time"] == 0
    assert df.loc[2, "origin_time"] == 3
    assert df.loc[3, "origin_time"] == 3


@pytest.mark.parametrize("df_type", [pd.DataFrame, pl.DataFrame])
def test_assign_trie_origin_times_node_rank_complex(df_type: typing.Type):
    df = df_type(
        {
            "id": [0, 1, 2, 3, 4, 5, 6, 7],
            "ancestor_id": [0, 0, 1, 2, 1, 2, 3, 3],
            "rank": [0, 0, 4, 7, 4, 7, 10, 10],
            "taxon_label": [None, None, "A", "B", "C", "D", "E", "F"],
            "differentia": [None, 0, 10, 7, 1, None, None, None],
        },
    )
    df = hstrat.AssignOriginTimeNodeRankTriePostprocessor(t0=-1)(
        df, p_differentia_collision=0.5, mutate=True
    )
    df = coerce_to_pandas(df)
    assert df.loc[0, "origin_time"] == 0 + 1
    assert df.loc[1, "origin_time"] == 0 + 1
    assert df.loc[2, "origin_time"] == 4 + 1
    assert df.loc[3, "origin_time"] == 7 + 1
    assert df.loc[4, "origin_time"] == 4 + 1
    assert df.loc[5, "origin_time"] == 7 + 1
    assert df.loc[6, "origin_time"] == 10 + 1
    assert df.loc[7, "origin_time"] == 10 + 1


@pytest.mark.parametrize("df_type", [pd.DataFrame, pl.DataFrame])
def test_assign_trie_origin_times_node_rank_assigned_property(
    df_type: typing.Type,
):
    df = pd.DataFrame(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 2],
            "rank": [0, 0, 0, 0],
            "taxon_label": [None, None, "A", "B"],
            "differentia": [None, 0, None, None],
        },
    )
    df = hstrat.AssignOriginTimeNodeRankTriePostprocessor(
        assigned_property="blueberry",
    )(df, p_differentia_collision=0.5, mutate=True)
    assert df.loc[0, "blueberry"] == 0
    assert df.loc[1, "blueberry"] == 0
    assert df.loc[2, "blueberry"] == 0


@pytest.mark.parametrize("df_type", [pd.DataFrame, pl.DataFrame])
def test_assign_trie_origin_times_node_rank_mutate(df_type: typing.Type):
    df = df_type(
        {
            "id": [0, 1, 2, 3],
            "ancestor_id": [0, 0, 1, 2],
            "rank": [0, 0, 0, 0],
            "taxon_label": [None, None, "A", "B"],
            "differentia": [None, 0, None, None],
        },
    )
    df_ = hstrat.AssignOriginTimeNodeRankTriePostprocessor()(
        df, p_differentia_collision=0.5, mutate=False
    )
    assert "origin_time" in df_.columns
    assert "origin_time" not in df.columns
    df = hstrat.AssignOriginTimeNodeRankTriePostprocessor(
        assigned_property="blueberry",
    )(df, p_differentia_collision=0.5, mutate=True)
    assert "blueberry" in df.columns



"""Provides numba TypingError exeption or tuple as exception nop.

Type must be declared outside jit'ed function or numba fails.
"""

from ._is_in_coverage_run import is_in_coverage_run

try:
    import numba as nb
except (ImportError, ModuleNotFoundError):  # pragma: no cover
    jit_TypingError = tuple()  # empty tuple is nop in except clause
else:
    if is_in_coverage_run():
        # when numba disabled for coverage testing,
        jit_TypingError = tuple()  # empty tuple is nop in except clause
    else:  # pragma: no cover
        # exclude from coverage because jit compilation disabled in cov runs
        jit_TypingError = nb.TypingError



import pickle
import tempfile

from hstrat.hstrat import perfect_resolution_algo


def test_init():
    assert (
        perfect_resolution_algo.Policy().GetSpec()
        == perfect_resolution_algo.Policy(
            policy_spec=perfect_resolution_algo.PolicySpec(),
        ).GetSpec()
    )

    policy = perfect_resolution_algo.Policy()

    # invariants
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBound)
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyAbsUpperBoundPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBound)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank)
    assert callable(policy.CalcMrcaUncertaintyRelUpperBoundPessimalRank)
    assert callable(policy.CalcNumStrataRetainedUpperBound)
    # scrying
    assert callable(policy.CalcMrcaUncertaintyAbsExact)
    assert callable(policy.CalcMrcaUncertaintyRelExact)
    assert callable(policy.CalcNumStrataRetainedExact)
    assert callable(policy.CalcRankAtColumnIndex)
    assert callable(policy.IterRetainedRanks)
    # enactment
    assert callable(policy.GenDropRanks)


def test_eq():
    policy = perfect_resolution_algo.Policy()
    assert policy == policy
    assert policy == perfect_resolution_algo.Policy()
    assert policy != policy.WithoutCalcRankAtColumnIndex()
    assert (
        policy.WithoutCalcRankAtColumnIndex()
        == policy.WithoutCalcRankAtColumnIndex()
    )


def test_pickle():
    original = perfect_resolution_algo.Policy()
    with tempfile.TemporaryDirectory() as tmp_path:
        with open(f"{tmp_path}/data", "wb") as tmp_file:
            pickle.dump(original, tmp_file)

        with open(f"{tmp_path}/data", "rb") as tmp_file:
            reconstituted = pickle.load(tmp_file)
            assert reconstituted == original


def test_GetSpec():
    assert perfect_resolution_algo.Policy().GetSpec()


def test_WithoutCalcRankAtColumnIndex():

    original = perfect_resolution_algo.Policy()
    stripped = original.WithoutCalcRankAtColumnIndex()

    assert stripped.CalcRankAtColumnIndex is None

    assert (
        original.CalcMrcaUncertaintyAbsUpperBound
        == stripped.CalcMrcaUncertaintyAbsUpperBound
    )
    assert (
        original.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank
        == stripped.CalcMrcaUncertaintyAbsUpperBoundAtPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyAbsUpperBoundPessimalRank
        == stripped.CalcMrcaUncertaintyAbsUpperBoundPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBound
        == stripped.CalcMrcaUncertaintyRelUpperBound
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank
        == stripped.CalcMrcaUncertaintyRelUpperBoundAtPessimalRank
    )
    assert (
        original.CalcMrcaUncertaintyRelUpperBoundPessimalRank
        == stripped.CalcMrcaUncertaintyRelUpperBoundPessimalRank
    )
    assert (
        original.CalcNumStrataRetainedUpperBound
        == stripped.CalcNumStrataRetainedUpperBound
    )
    # scrying
    assert (
        original.CalcMrcaUncertaintyAbsExact
        == stripped.CalcMrcaUncertaintyAbsExact
    )
    assert (
        original.CalcMrcaUncertaintyRelExact
        == stripped.CalcMrcaUncertaintyRelExact
    )
    assert (
        original.CalcNumStrataRetainedExact
        == stripped.CalcNumStrataRetainedExact
    )
    assert original.IterRetainedRanks == stripped.IterRetainedRanks
    # enactment
    assert original.GenDropRanks == stripped.GenDropRanks

    # test chaining
    assert (
        perfect_resolution_algo.Policy().WithoutCalcRankAtColumnIndex()
        == stripped
    )


def test_repr():
    policy = perfect_resolution_algo.Policy()
    assert policy.GetSpec().GetAlgoIdentifier() in repr(policy)


def test_str():
    policy = perfect_resolution_algo.Policy()
    assert policy.GetSpec().GetAlgoTitle() in str(policy)



from hstrat._auxiliary_lib import join_paragraphs_from_one_sentence_per_line


def test_single_sentence_per_line():
    text = "This is a sentence.\nThis is another one."
    expected = "This is a sentence. This is another one."
    assert join_paragraphs_from_one_sentence_per_line(text) == expected


def test_multiple_paragraphs():
    text = "First paragraph sentence one.\nFirst paragraph sentence two.\n\nSecond paragraph sentence one.\nSecond paragraph sentence two."
    expected = "First paragraph sentence one. First paragraph sentence two. \nSecond paragraph sentence one. Second paragraph sentence two."
    assert join_paragraphs_from_one_sentence_per_line(text) == expected


def test_no_newlines():
    text = "This is a sentence. This is another one."
    expected = "This is a sentence. This is another one."
    assert join_paragraphs_from_one_sentence_per_line(text) == expected


def test_empty_input():
    text = ""
    expected = ""
    assert join_paragraphs_from_one_sentence_per_line(text) == expected


def test_newline_at_end():
    text = "This is a sentence.\n"
    expected = "This is a sentence. "
    assert join_paragraphs_from_one_sentence_per_line(text) == expected



import typing

import opytional as opyt

from ..._auxiliary_lib import (
    BioPhyloTree,
    HereditaryStratigraphicArtifact,
    to_tril,
)
from ._build_distance_matrix_numpy import build_distance_matrix_numpy


def build_distance_matrix_biopython(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    estimator: str,
    prior: typing.Union[str, typing.Any],
    taxon_labels: typing.Optional[typing.Iterable] = None,
    force_common_ancestry: typing.Optional[bool] = False,
) -> BioPhyloTree.DistanceMatrix:
    """Compute all-pairs patristic distance among a population of extant
    hereditary stratigraphic columns as a BioPython `DistanceMatrix`.

    Parameters
    ----------
    population : Sequence[HereditaryStratigraphicArtifact]
        The extant hereditary stratigraphic columns to compare.

        The ordering of rows and columns within the returned matrix will
        correspond to the ordering of columns in this sequence.
    estimator : {"maximum_likelihood", "unbiased"}
        What patristic distance estimation method should be used? Options are
        "maximum_likelihood" or "unbiased".

        See `estimate_ranks_since_mrca_with` for discussion of estimator
        options.
    prior :
        Prior probability density distribution over possible generations of the
        MRCA for MRCA estimation.

        See `estimate_ranks_since_mrca_with` for discussion of prior
        options.
    taxon_labels : Iterable[str]], optional
        How should leaf nodes representing extant hereditary stratigraphic
        columns be named?

        Label order should correspond to the order of corresponding hereditary
        stratigraphic columns within `population`. If None, taxons will be
        named according to their numerical index.
    force_common_ancestry : Optional[bool], optional
        How should columns that definitively share no common ancestry be handled?

        If set to True, treat columns with no common ancestry as if they
        shared a common ancestor immediately before the genesis of the
        lineages. If set to False, set the patristic distance between columns
        with no common ancestry as NaN.

        If set to None, as default, the presence of columns that definitively
        share no common ancestry will raise a ValueError.

    Returns
    -------
    np.ndarray
        The patristic distance matrix of the population, a square numpy array
        of shape (len(population), len(population)).

    Raises
    ------
    ValueError
        If the distance between two columns that definitively share no common
        ancestry are attempted to be computed without setting
        `force_common_ancestry`.
    """

    distance_matrix = build_distance_matrix_numpy(
        population,
        estimator,
        prior,
        force_common_ancestry,
    )
    return BioPhyloTree.DistanceMatrix(
        names=opyt.or_value(
            opyt.apply_if(taxon_labels, list),
            [*map(str, range(len(population)))],
        ),
        matrix=to_tril(distance_matrix),
    )



import itertools as it
import random

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.95, 0.88],
)
def test_calc_rank_of_earliest_detectable_mrca_between_specimen(
    retention_policy, differentia_width, confidence_level
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(100)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()

    assert hstrat.calc_rank_of_earliest_detectable_mrca_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column2),
        confidence_level=confidence_level,
    ) == hstrat.calc_rank_of_earliest_detectable_mrca_between(
        column, column2, confidence_level=confidence_level
    )
    assert hstrat.calc_rank_of_earliest_detectable_mrca_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(column),
        confidence_level=confidence_level,
    ) == hstrat.calc_rank_of_earliest_detectable_mrca_between(
        column, column, confidence_level=confidence_level
    )

    assert hstrat.calc_rank_of_earliest_detectable_mrca_between(
        hstrat.col_to_specimen(column),
        hstrat.col_to_specimen(child1),
        confidence_level=confidence_level,
    ) == hstrat.calc_rank_of_earliest_detectable_mrca_between(
        column, child1, confidence_level=confidence_level
    )

    assert hstrat.calc_rank_of_earliest_detectable_mrca_between(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        confidence_level=confidence_level,
    ) == hstrat.calc_rank_of_earliest_detectable_mrca_between(
        child1, child2, confidence_level=confidence_level
    )

    child1.DepositStrata(10)
    assert hstrat.calc_rank_of_earliest_detectable_mrca_between(
        hstrat.col_to_specimen(child1),
        hstrat.col_to_specimen(child2),
        confidence_level=confidence_level,
    ) == hstrat.calc_rank_of_earliest_detectable_mrca_between(
        child1, child2, confidence_level=confidence_level
    )


@pytest.mark.filterwarnings(
    "ignore:Insufficient common ranks between columns to detect common ancestry at given confidence level."
)
@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95, 0.99],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_CalcRankOfEarliestDetectableMrcaWith1(
    confidence_level,
    differentia_bit_width,
    wrap,
):
    expected_thresh = (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_bit_width,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1 - confidence_level,
        )
        - 1
    )

    for r1, r2, r3 in it.product(*[range(1, expected_thresh)] * 3):
        c1 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_bit_width,
            stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
        )
        c2 = c1.Clone()
        c3 = hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_bit_width,
            stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
        )
        for __ in range(r1 - 1):
            c1.DepositStratum()
        for __ in range(r2 - 1):
            c2.DepositStratum()
        for __ in range(r3 - 1):
            c3.DepositStratum()

        for x1, x2 in it.combinations([c1, c2, c3], 2):
            assert (
                hstrat.calc_rank_of_earliest_detectable_mrca_between(
                    wrap(x1),
                    wrap(x2),
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_earliest_detectable_mrca_between(
                    wrap(x2),
                    wrap(x1),
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_mrca_bounds_between(
                    x1,
                    x2,
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )
            assert (
                hstrat.calc_rank_of_mrca_bounds_between(
                    x2,
                    x1,
                    prior="arbitrary",
                    confidence_level=confidence_level,
                )
                is None
            )


@pytest.mark.parametrize(
    "confidence_level",
    [0.8, 0.95, 0.99],
)
@pytest.mark.parametrize(
    "differentia_bit_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_CalcRankOfEarliestDetectableMrcaWith2(
    confidence_level,
    differentia_bit_width,
    wrap,
):
    expected_thresh = (
        hstrat.HereditaryStratigraphicColumn(
            stratum_differentia_bit_width=differentia_bit_width,
        ).CalcMinImplausibleSpuriousConsecutiveDifferentiaCollisions(
            significance_level=1 - confidence_level,
        )
        - 1
    )

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_bit_width,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    c2 = c1.Clone()
    c3 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=differentia_bit_width,
        stratum_retention_policy=hstrat.fixed_resolution_algo.Policy(2),
    )

    for x1, x2 in it.combinations([c1, c2, c3], 2):
        x1 = x1.Clone()
        x2 = x2.Clone()
        while (
            hstrat.get_nth_common_rank_between(x1, x2, expected_thresh) is None
        ):
            random.choice([x1, x2]).DepositStratum()

        assert hstrat.calc_rank_of_earliest_detectable_mrca_between(
            wrap(x1),
            wrap(x2),
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x1, x2, expected_thresh)
        assert hstrat.calc_rank_of_earliest_detectable_mrca_between(
            wrap(x2),
            wrap(x1),
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x2, x1, expected_thresh)

        for __ in range(3):
            x1.DepositStratum()
            x2.DepositStratum()

        assert hstrat.calc_rank_of_earliest_detectable_mrca_between(
            wrap(x1),
            wrap(x2),
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x1, x2, expected_thresh)
        assert hstrat.calc_rank_of_earliest_detectable_mrca_between(
            wrap(x2),
            wrap(x1),
            confidence_level=confidence_level,
        ) == hstrat.get_nth_common_rank_between(x2, x1, expected_thresh)


@pytest.mark.parametrize(
    "wrap",
    [
        lambda x: x,
        hstrat.col_to_specimen,
    ],
)
def test_CalcRankOfEarliestDetectableMrcaWith3(wrap):

    c1 = hstrat.HereditaryStratigraphicColumn(
        stratum_differentia_bit_width=64,
        stratum_retention_policy=hstrat.perfect_resolution_algo.Policy(),
    )
    for __ in range(10):
        assert (
            hstrat.calc_rank_of_earliest_detectable_mrca_between(
                wrap(c1), wrap(c1)
            )
            == 0
        )
        c1.DepositStratum()


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
def test_artifact_types_equiv(differentia_width, policy):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.calc_rank_of_earliest_detectable_mrca_between(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
        ) == hstrat.calc_rank_of_earliest_detectable_mrca_between(a, b)



import itertools as it

import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
        hstrat.recency_proportional_resolution_algo.Policy(
            recency_proportional_resolution=2
        ),
    ],
)
@pytest.mark.parametrize(
    "differentia_width",
    [1, 2, 8, 64],
)
@pytest.mark.parametrize(
    "wrap",
    [lambda x: x, hstrat.col_to_specimen],
)
def test_ballpark_rank_of_mrca_between(
    retention_policy, differentia_width, wrap
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column2 = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
        stratum_differentia_bit_width=differentia_width,
    )
    column.DepositStrata(25)

    child1 = column.CloneDescendant()
    child2 = column.CloneDescendant()
    grandchild = child1.CloneNthDescendant(20)

    for first, second in it.product(
        *[[column, column2, child1, child2, grandchild]] * 2
    ):
        assert hstrat.estimate_rank_of_mrca_between(
            wrap(first),
            wrap(second),
            estimator="maximum_likelihood",
            prior="arbitrary",
        ) == hstrat.ballpark_rank_of_mrca_between(first, second)


@pytest.mark.parametrize(
    "differentia_width",
    [1, 8, 64],
)
@pytest.mark.parametrize(
    "policy",
    [
        hstrat.fixed_resolution_algo.Policy(3),
        hstrat.recency_proportional_resolution_algo.Policy(1),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.perfect_resolution_algo.Policy(),
    ],
)
@pytest.mark.parametrize(
    "prior",
    ["arbitrary"],
)
def test_artifact_types_equiv(differentia_width, policy, prior):
    common_ancestor = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c1 = common_ancestor.CloneNthDescendant(4)
    c2 = common_ancestor.CloneNthDescendant(9)
    c_x = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    ).CloneNthDescendant(7)
    c_y = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=policy,
        stratum_differentia_bit_width=differentia_width,
    )

    for a, b in it.product(
        [common_ancestor, c1, c2, c_x, c_y],
        [common_ancestor, c1, c2, c_x, c_y],
    ):
        assert hstrat.ballpark_rank_of_mrca_between(
            hstrat.col_to_specimen(a),
            hstrat.col_to_specimen(b),
        ) == hstrat.ballpark_rank_of_mrca_between(a, b)



from dataframe import surface_postprocess_trie, surface_unpack_reconstruct
from frozen_instrumentation import (
    HereditaryStratigraphicAssemblage,
    HereditaryStratigraphicAssemblageSpecimen,
    HereditaryStratigraphicSpecimen,
)
from genome_instrumentation import (
    HereditaryStratigraphicColumn,
    HereditaryStratigraphicColumnBundle,
    HereditaryStratum,
    HereditaryStratumOrderedStoreDict,
    HereditaryStratumOrderedStoreList,
    HereditaryStratumOrderedStoreTree,
    provided_stratum_ordered_stores,
    stratum_ordered_stores,
)
from juxtaposition import (
    calc_definitive_max_rank_of_first_retained_disparity_between,
    calc_definitive_max_rank_of_last_retained_commonality_between,
    calc_definitive_min_ranks_since_first_retained_disparity_with,
    calc_definitive_min_ranks_since_last_retained_commonality_with,
    calc_min_implausible_spurious_consecutive_differentia_collisions_between,
    calc_probability_differentia_collision_between,
    calc_rank_of_first_retained_disparity_between,
    calc_rank_of_last_retained_commonality_between,
    calc_ranks_since_first_retained_disparity_with,
    calc_ranks_since_last_retained_commonality_with,
    diff_retained_ranks,
    get_last_common_stratum_between,
    get_nth_common_rank_between,
)
from phylogenetic_inference import (
    ArbitraryPrior,
    AssignDestructionTimeYoungestPlusOneTriePostprocessor,
    AssignOriginTimeExpectedValueTriePostprocessor,
    AssignOriginTimeNaiveTriePostprocessor,
    AssignOriginTimeNodeRankTriePostprocessor,
    AssignOriginTimeSampleNaiveTriePostprocessor,
    CompoundTriePostprocessor,
    ExponentialPrior,
    GeometricPrior,
    NopTriePostprocessor,
    PeelBackConjoinedLeavesTriePostprocessor,
    SampleAncestralRollbacksTriePostprocessor,
    UniformPrior,
    ballpark_patristic_distance_between,
    ballpark_rank_of_mrca_between,
    ballpark_ranks_since_mrca_with,
    build_distance_matrix_biopython,
    build_distance_matrix_numpy,
    build_tree,
    build_tree_nj,
    build_tree_searchtable,
    build_tree_trie,
    build_tree_trie_ensemble,
    build_tree_upgma,
    calc_patristic_distance_bounds_between,
    calc_rank_of_earliest_detectable_mrca_among,
    calc_rank_of_earliest_detectable_mrca_between,
    calc_rank_of_mrca_bounds_among,
    calc_rank_of_mrca_bounds_between,
    calc_rank_of_mrca_bounds_provided_confidence_level,
    calc_rank_of_mrca_uncertainty_among,
    calc_rank_of_mrca_uncertainty_between,
    calc_ranks_since_earliest_detectable_mrca_with,
    calc_ranks_since_mrca_bounds_provided_confidence_level,
    calc_ranks_since_mrca_bounds_with,
    calc_ranks_since_mrca_uncertainty_with,
    does_definitively_have_no_common_ancestor,
    does_definitively_share_no_common_ancestor,
    does_have_any_common_ancestor,
    does_share_any_common_ancestor,
    estimate_patristic_distance_between,
    estimate_rank_of_mrca_between,
    estimate_rank_of_mrca_maximum_likelihood,
    estimate_rank_of_mrca_naive,
    estimate_rank_of_mrca_unbiased,
    estimate_ranks_since_mrca_with,
    estimators,
    pairwise,
    population,
    priors,
    tree,
    trie_postprocess,
)
from serialization import (
    assemblage_from_records,
    col_from_int,
    col_from_packet,
    col_from_packet_buffer,
    col_from_records,
    col_to_dataframe,
    col_to_int,
    col_to_packet,
    col_to_records,
    col_to_specimen,
    pack_differentiae,
    pack_differentiae_bytes,
    pack_differentiae_str,
    policy_from_records,
    policy_to_records,
    pop_from_records,
    pop_to_assemblage,
    pop_to_dataframe,
    pop_to_records,
    specimen_from_records,
    unassemblage_from_records,
    unpack_differentiae,
    unpack_differentiae_bytes,
    unpack_differentiae_str,
)
from stratum_retention_strategy import (
    MrcaUncertaintyAbsExactEvaluator,
    MrcaUncertaintyAbsUpperBoundEvaluator,
    MrcaUncertaintyRelExactEvaluator,
    MrcaUncertaintyRelUpperBoundEvaluator,
    NumStrataRetainedExactEvaluator,
    NumStrataRetainedUpperBoundEvaluator,
    PropertyAtLeastParameterizer,
    PropertyAtMostParameterizer,
    PropertyExactlyParameterizer,
    UnsatisfiableParameterizationRequestError,
    depth_proportional_resolution_algo,
    depth_proportional_resolution_tapered_algo,
    fixed_resolution_algo,
    geom_seq_nth_root_algo,
    geom_seq_nth_root_tapered_algo,
    nominal_resolution_algo,
    perfect_resolution_algo,
    provided_stratum_retention_algorithms,
    pseudostochastic_algo,
    recency_proportional_resolution_algo,
    recency_proportional_resolution_curbed_algo,
    stochastic_algo,
    stratum_retention_algorithms,
    stratum_retention_policy_evaluators,
    stratum_retention_policy_parameterizers,
)
from stratum_retention_viz import (
    animate,
    ascii,
    col_to_ascii,
    mrca_uncertainty_absolute_barplot,
    mrca_uncertainty_relative_barplot,
    plot,
    policy_panel_animate,
    policy_panel_plot,
    strata_retained_frac_lineplot,
    strata_retained_num_lineplot,
    stratum_retention_animate,
    stratum_retention_dripplot,
)
from test_drive import (
    DecantingPhyloTracker,
    GarbageCollectingPhyloTracker,
    PerfectBacktrackHandle,
    compile_perfect_backtrack_phylogeny,
    compile_phylogeny_from_lineage_iters,
    descend_template_phylogeny,
    descend_template_phylogeny_,
    descend_template_phylogeny_alifestd,
    descend_template_phylogeny_biopython,
    descend_template_phylogeny_dendropy,
    descend_template_phylogeny_naive,
    descend_template_phylogeny_networkx,
    descend_template_phylogeny_posthoc,
    evolve_fitness_trait_population,
    generate_template_phylogeny,
    iter_perfect_backtrack_lineage,
    perfect_tracking,
)

__all__ = [
    # dataframe
    "surface_postprocess_trie",
    "surface_unpack_reconstruct",
    # frozen_instrumentation
    "HereditaryStratigraphicAssemblage",
    "HereditaryStratigraphicAssemblageSpecimen",
    "HereditaryStratigraphicSpecimen",
    # genome_instrumentation
    "stratum_ordered_stores",
    "HereditaryStratum",
    "HereditaryStratigraphicColumn",
    "HereditaryStratigraphicColumnBundle",
    "HereditaryStratumOrderedStoreDict",
    "HereditaryStratumOrderedStoreList",
    "HereditaryStratumOrderedStoreTree",
    "provided_stratum_ordered_stores",
    # juxtaposition
    "calc_definitive_max_rank_of_first_retained_disparity_between",
    "calc_definitive_max_rank_of_last_retained_commonality_between",
    "calc_definitive_min_ranks_since_first_retained_disparity_with",
    "calc_definitive_min_ranks_since_last_retained_commonality_with",
    "calc_min_implausible_spurious_consecutive_differentia_collisions_between",
    "calc_probability_differentia_collision_between",
    "calc_rank_of_first_retained_disparity_between",
    "calc_rank_of_last_retained_commonality_between",
    "calc_ranks_since_first_retained_disparity_with",
    "calc_ranks_since_last_retained_commonality_with",
    "diff_retained_ranks",
    "get_last_common_stratum_between",
    "get_nth_common_rank_between",
    # phylogenetic_inference
    "estimators",
    "pairwise",
    "population",
    "priors",
    "tree",
    "estimate_rank_of_mrca_maximum_likelihood",
    "estimate_rank_of_mrca_naive",
    "estimate_rank_of_mrca_unbiased",
    "ballpark_patristic_distance_between",
    "ballpark_rank_of_mrca_between",
    "ballpark_ranks_since_mrca_with",
    "calc_patristic_distance_bounds_between",
    "calc_rank_of_earliest_detectable_mrca_between",
    "calc_rank_of_mrca_bounds_between",
    "calc_rank_of_mrca_bounds_provided_confidence_level",
    "calc_rank_of_mrca_uncertainty_between",
    "calc_ranks_since_earliest_detectable_mrca_with",
    "calc_ranks_since_mrca_bounds_with",
    "calc_ranks_since_mrca_bounds_provided_confidence_level",
    "calc_ranks_since_mrca_uncertainty_with",
    "does_definitively_have_no_common_ancestor",
    "does_have_any_common_ancestor",
    "estimate_patristic_distance_between",
    "estimate_rank_of_mrca_between",
    "estimate_ranks_since_mrca_with",
    "build_distance_matrix_biopython",
    "build_distance_matrix_numpy",
    "calc_rank_of_earliest_detectable_mrca_among",
    "calc_rank_of_mrca_bounds_among",
    "calc_rank_of_mrca_uncertainty_among",
    "does_definitively_share_no_common_ancestor",
    "does_share_any_common_ancestor",
    "ArbitraryPrior",
    "GeometricPrior",
    "ExponentialPrior",
    "UniformPrior",
    "build_tree_nj",
    "build_tree_searchtable",
    "build_tree_trie",
    "build_tree_trie_ensemble",
    "build_tree_upgma",
    "build_tree",
    "trie_postprocess",
    "AssignDestructionTimeYoungestPlusOneTriePostprocessor",
    "AssignOriginTimeExpectedValueTriePostprocessor",
    "AssignOriginTimeNaiveTriePostprocessor",
    "AssignOriginTimeNodeRankTriePostprocessor",
    "AssignOriginTimeSampleNaiveTriePostprocessor",
    "CompoundTriePostprocessor",
    "NopTriePostprocessor",
    "PeelBackConjoinedLeavesTriePostprocessor",
    "SampleAncestralRollbacksTriePostprocessor",
    # serialization
    "assemblage_from_records",
    "col_from_int",
    "col_from_packet",
    "col_from_packet_buffer",
    "col_from_records",
    "col_to_dataframe",
    "col_to_int",
    "col_to_packet",
    "col_to_records",
    "col_to_specimen",
    "pack_differentiae",
    "pack_differentiae_bytes",
    "pack_differentiae_str",
    "policy_from_records",
    "policy_to_records",
    "pop_from_records",
    "pop_to_assemblage",
    "pop_to_dataframe",
    "pop_to_records",
    "specimen_from_records",
    "unassemblage_from_records",
    "unpack_differentiae",
    "unpack_differentiae_bytes",
    "unpack_differentiae_str",
    # stratum_retention_strategy
    "stratum_retention_policy_parameterizers",
    "stratum_retention_algorithms",
    "stratum_retention_policy_evaluators",
    "MrcaUncertaintyAbsExactEvaluator",
    "MrcaUncertaintyAbsUpperBoundEvaluator",
    "MrcaUncertaintyRelExactEvaluator",
    "MrcaUncertaintyRelUpperBoundEvaluator",
    "NumStrataRetainedExactEvaluator",
    "NumStrataRetainedUpperBoundEvaluator",
    "UnsatisfiableParameterizationRequestError",
    "depth_proportional_resolution_algo",
    "depth_proportional_resolution_tapered_algo",
    "fixed_resolution_algo",
    "geom_seq_nth_root_algo",
    "geom_seq_nth_root_tapered_algo",
    "nominal_resolution_algo",
    "perfect_resolution_algo",
    "provided_stratum_retention_algorithms",
    "pseudostochastic_algo",
    "recency_proportional_resolution_algo",
    "recency_proportional_resolution_curbed_algo",
    "stochastic_algo",
    "PropertyAtLeastParameterizer",
    "PropertyAtMostParameterizer",
    "PropertyExactlyParameterizer",
    # stratum_retention_viz
    "animate",
    "ascii",
    "plot",
    "col_to_ascii",
    "stratum_retention_animate",
    "policy_panel_animate",
    "mrca_uncertainty_absolute_barplot",
    "mrca_uncertainty_relative_barplot",
    "policy_panel_plot",
    "strata_retained_frac_lineplot",
    "strata_retained_num_lineplot",
    "stratum_retention_dripplot",
    # test_drive
    "descend_template_phylogeny_",
    "generate_template_phylogeny",
    "perfect_tracking",
    "descend_template_phylogeny",
    "descend_template_phylogeny_alifestd",
    "descend_template_phylogeny_biopython",
    "descend_template_phylogeny_dendropy",
    "descend_template_phylogeny_naive",
    "descend_template_phylogeny_networkx",
    "descend_template_phylogeny_posthoc",
    "evolve_fitness_trait_population",
    "compile_perfect_backtrack_phylogeny",
    "compile_phylogeny_from_lineage_iters",
    "DecantingPhyloTracker",
    "GarbageCollectingPhyloTracker",
    "iter_perfect_backtrack_lineage",
    "PerfectBacktrackHandle",
]



import typing

from ..genome_instrumentation import HereditaryStratum
from ._impl import stringify_packed_differentia_bytes
from ._pack_differentiae_bytes import pack_differentiae_bytes


def pack_differentiae_str(
    strata: typing.Iterable[HereditaryStratum],
    differentia_bit_width: int,
) -> str:
    """Pack a sequence of differentiae together into a compact string
    representation.

    Returns a string with base 64 encoded concatenation of diffferentiae.
    If `differentia_bit_width` is not an even byte multiple, the first encoded
    byte tells how many empty padding bits, if any, were placed at the end of
    the concatenation in order to align the bitstring end to byte boundaries.
    """

    buffer = pack_differentiae_bytes(strata, differentia_bit_width)
    return stringify_packed_differentia_bytes(buffer)



import anytree
import pytest

from hstrat._auxiliary_lib import (
    anytree_cardinality,
    anytree_peel_sibling_to_cousin,
    is_in,
)


def test_anytree_peel_sibling_to_cousin_with_leaf_node():
    #    A
    #   / \
    #  B   C
    # /
    # D
    root = anytree.Node("A")
    b = anytree.Node("B", parent=root)
    anytree.Node("C", parent=root)
    d = anytree.Node("D", parent=b)

    with pytest.raises(AssertionError):
        anytree_peel_sibling_to_cousin(d)

    # Tree structure should not change
    #    A
    #   / \
    #  B   C
    # /
    # D


def test_anytree_peel_sibling_to_cousin_with_node_without_sibling():
    #    A
    #   / \
    #  B   C
    root = anytree.Node("A")
    anytree.Node("B", parent=root)
    anytree.Node("C", parent=root)

    with pytest.raises(AssertionError):
        anytree_peel_sibling_to_cousin(root)

    # Tree structure should not change
    #    A
    #   / \
    #  B   C


def test_anytree_peel_sibling_to_cousin_with_node_with_sibling():
    #    A
    #   / \
    #  B   C
    # / \
    # D   E
    #    / \
    #   F   G
    a = anytree.Node("A")
    b = anytree.Node("B", parent=a)
    c = anytree.Node("C", parent=a)
    d = anytree.Node("D", parent=b)
    e = anytree.Node("E", parent=b)
    f = anytree.Node("F", parent=e)
    g = anytree.Node("G", parent=e)
    assert anytree_cardinality(a) == 7

    anytree_peel_sibling_to_cousin(e)
    # Tree structure after the first peel:
    #    A
    #   /|\
    #  B C B'
    #  |   |
    #  D   E
    #     / \
    #    F   G

    b_prime = next(node for node in a.children if not is_in(node, (b, c)))
    assert b_prime.name == "B"

    assert anytree_cardinality(a) == 8
    assert a.parent is None
    assert b.parent is a
    assert b_prime.parent is a
    assert c.parent is a
    assert d.parent is b
    assert e.parent is b_prime
    assert f.parent is e
    assert g.parent is e

    anytree_peel_sibling_to_cousin(g)
    # Tree structure after the second peel:
    #    A
    #   /|\
    #  B C B'
    #  |  / \
    #  D  E  E'
    #     |  |
    #     F  G

    e_prime = next(node for node in b_prime.children if node is not e)
    assert e_prime.name == "E"

    assert anytree_cardinality(a) == 9
    assert a.parent is None
    assert b.parent is a
    assert b_prime.parent is a
    assert c.parent is a
    assert d.parent is b
    assert e.parent is b_prime
    assert e_prime.parent is b_prime
    assert f.parent is e
    assert g.parent is e_prime

    anytree_peel_sibling_to_cousin(e_prime)
    # Tree structure after the third peel:
    #      A
    #   /|\  \
    #  B C B' B''
    #  |   |  |
    #  D   E  E'
    #      |  |
    #      F  G

    b_double_prime = next(
        node for node in a.children if not is_in(node, (b, c, b_prime))
    )
    assert b_double_prime.name == "B"

    assert anytree_cardinality(a) == 10
    assert a.parent is None
    assert b.parent is a
    assert b_prime.parent is a
    assert b_double_prime.parent is a
    assert c.parent is a
    assert d.parent is b
    assert e.parent is b_prime
    assert e_prime.parent is b_double_prime
    assert f.parent is e
    assert g.parent is e_prime



import pytest

from hstrat import hstrat


@pytest.mark.parametrize(
    "retention_policy",
    [
        hstrat.perfect_resolution_algo.Policy(),
        hstrat.nominal_resolution_algo.Policy(),
        hstrat.fixed_resolution_algo.Policy(fixed_resolution=10),
    ],
)
def test_col_to_dataframe(
    retention_policy,
):
    column = hstrat.HereditaryStratigraphicColumn(
        stratum_retention_policy=retention_policy,
    )

    for generation in range(100):
        column.DepositStratum()

    df = hstrat.col_to_dataframe(column)
    assert len(df) == column.GetNumStrataRetained()
    assert df["rank"].max() == column.GetNumStrataDeposited() - 1



import unittest

from hstrat import hstrat
from hstrat._auxiliary_lib import is_strictly_increasing


class TestHereditaryStratumOrderedStoreList(unittest.TestCase):

    # tests can run independently
    _multiprocess_can_split_ = True

    def test_deposition(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        assert store1.GetNumStrataRetained() == 0

        stratum1 = hstrat.HereditaryStratum(deposition_rank=0)
        store1.DepositStratum(0, stratum1)
        assert store1.GetNumStrataRetained() == 1
        assert store1.GetStratumAtColumnIndex(0) == stratum1

        store2 = store1.Clone()

        stratum2 = hstrat.HereditaryStratum(deposition_rank=1)
        store1.DepositStratum(1, stratum2)
        assert store1.GetNumStrataRetained() == 2
        assert store1.GetStratumAtColumnIndex(1) == stratum2
        assert store1.GetStratumAtColumnIndex(0) != stratum2

        assert store2.GetNumStrataRetained() == 1
        assert store2.GetStratumAtColumnIndex(0) == stratum1

    def test_deletion1(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        stratum1 = hstrat.HereditaryStratum(deposition_rank=0)
        store1.DepositStratum(0, stratum1)

        store2 = store1.Clone()
        stratum2 = hstrat.HereditaryStratum(deposition_rank=1)
        store1.DepositStratum(1, stratum2)

        del store1
        assert store2.GetNumStrataRetained() == 1
        assert store2.GetStratumAtColumnIndex(0) == stratum1

    def test_deletion2(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        stratum1 = hstrat.HereditaryStratum(deposition_rank=0)
        store1.DepositStratum(0, stratum1)

        store2 = store1.Clone()
        stratum2 = hstrat.HereditaryStratum(deposition_rank=1)
        store1.DepositStratum(1, stratum2)

        del store2
        assert store1.GetNumStrataRetained() == 2
        assert store1.GetStratumAtColumnIndex(0) == stratum1
        assert store1.GetStratumAtColumnIndex(1) == stratum2

    def test_equality(self):
        assert (
            hstrat.HereditaryStratumOrderedStoreList()
            == hstrat.HereditaryStratumOrderedStoreList()
        )

        store1 = hstrat.HereditaryStratumOrderedStoreList()
        store1.DepositStratum(0, hstrat.HereditaryStratum(deposition_rank=0))
        store2 = store1.Clone()
        assert store1 == store2

        store2.DepositStratum(1, hstrat.HereditaryStratum(deposition_rank=1))
        assert store1 != store2

    def test_GetRankAtColumnIndex(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        store1.DepositStratum(0, hstrat.HereditaryStratum(deposition_rank=0))
        store1.DepositStratum(1, hstrat.HereditaryStratum(deposition_rank=1))
        store1.DepositStratum(2, hstrat.HereditaryStratum(deposition_rank=2))
        assert store1.GetRankAtColumnIndex(0) == 0
        assert store1.GetRankAtColumnIndex(1) == 1
        assert store1.GetRankAtColumnIndex(2) == 2

        store1.DelRanks([1])
        assert store1.GetRankAtColumnIndex(0) == 0
        assert store1.GetRankAtColumnIndex(1) == 2

    def test_GetStratumAtColumnIndex(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        strata = [
            hstrat.HereditaryStratum(deposition_rank=rank) for rank in range(3)
        ]
        for rank, stratum in enumerate(strata):
            store1.DepositStratum(rank, stratum)

        for rank, stratum in enumerate(strata):
            assert store1.GetStratumAtColumnIndex(rank) == strata[rank]

        store1.DelRanks([1])
        assert store1.GetStratumAtColumnIndex(0) == strata[0]
        assert store1.GetStratumAtColumnIndex(1) == strata[2]

    def test_GetNumStrataRetained(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        for rank in range(5):
            assert store1.GetNumStrataRetained() == rank
            store1.DepositStratum(rank, hstrat.HereditaryStratum())
        assert store1.GetNumStrataRetained() == 5

        store1.DelRanks([1, 2], get_column_index_of_rank=lambda x: x)
        assert store1.GetNumStrataRetained() == 3

        store1.DepositStratum(5, hstrat.HereditaryStratum())
        assert store1.GetNumStrataRetained() == 4

        store1.DelRanks(
            [5],
            get_column_index_of_rank=lambda x: {
                3: 0,
                5: 1,
            }[x],
        )
        assert store1.GetNumStrataRetained() == 3

    def test_GetColumnIndexOfRank(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 63]
        for rank in ranks:
            store1.DepositStratum(
                rank=rank,
                stratum=hstrat.HereditaryStratum(deposition_rank=rank),
            )

        assert store1.GetColumnIndexOfRank(-1) is None
        assert store1.GetColumnIndexOfRank(0) == 0
        assert store1.GetColumnIndexOfRank(1) is None
        assert store1.GetColumnIndexOfRank(8) == 1
        assert store1.GetColumnIndexOfRank(42) == 2
        assert store1.GetColumnIndexOfRank(63) == 3
        assert store1.GetColumnIndexOfRank(64) is None

    def test_IterRetainedRanks(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 63]
        for rank in ranks:
            store1.DepositStratum(
                rank=rank,
                stratum=hstrat.HereditaryStratum(deposition_rank=rank),
            )

        assert set(store1.IterRetainedRanks()) == set(ranks)
        assert [*store1.IterRetainedRanks()] == ranks
        assert is_strictly_increasing(ranks)

    def test_IterRankDifferentiaZip1(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 63]
        strata = [
            hstrat.HereditaryStratum(deposition_rank=rank) for rank in ranks
        ]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        assert [
            *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
        ] == [*store1.IterRankDifferentiaZip()]
        assert [*zip(ranks, [stratum.GetDifferentia() for stratum in strata])][
            0:
        ] == [*store1.IterRankDifferentiaZip(start_column_index=0)]
        assert [*zip(ranks, [stratum.GetDifferentia() for stratum in strata])][
            2:
        ] == [*store1.IterRankDifferentiaZip(start_column_index=2)]

    def test_IterRankDifferentiaZip2(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 63]
        strata = [
            hstrat.HereditaryStratum(deposition_rank=rank) for rank in ranks
        ]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        def col_index_to_rank(column_idx):
            return ranks[column_idx]

        assert [
            *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
        ] == [
            *store1.IterRankDifferentiaZip(
                get_rank_at_column_index=col_index_to_rank,
            )
        ]
        assert [*zip(ranks, [stratum.GetDifferentia() for stratum in strata])][
            0:
        ] == [
            *store1.IterRankDifferentiaZip(
                get_rank_at_column_index=col_index_to_rank,
                start_column_index=0,
            )
        ]
        assert [*zip(ranks, [stratum.GetDifferentia() for stratum in strata])][
            2:
        ] == [
            *store1.IterRankDifferentiaZip(
                get_rank_at_column_index=col_index_to_rank,
                start_column_index=2,
            )
        ]

    def test_DelRanks_getrank_impl1(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 55, 63]
        strata = [
            hstrat.HereditaryStratum(deposition_rank=rank) for rank in ranks
        ]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        for deletion in (
            [],
            [8, 42],
            [],
            [55],
            [],
            [0, 63],
            [],
        ):
            store1.DelRanks(deletion)
            for rank in deletion:
                del strata[ranks.index(rank)]
                ranks.remove(rank)
            assert [
                *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
            ] == [*store1.IterRankDifferentiaZip()]

    def test_DelRanks_getrank_impl2(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 55, 63]
        strata = [
            hstrat.HereditaryStratum(deposition_rank=rank) for rank in ranks
        ]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        for deletion in (
            [],
            [0, 63],
            [],
            [8],
            [],
            [55],
            [],
            [42],
            [],
        ):
            store1.DelRanks(deletion)
            for rank in deletion:
                del strata[ranks.index(rank)]
                ranks.remove(rank)
            assert [
                *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
            ] == [*store1.IterRankDifferentiaZip()]

    def test_DelRanks_getrank_impl3(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 55, 63, 80]
        strata = [
            hstrat.HereditaryStratum(deposition_rank=rank) for rank in ranks
        ]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        for deletion in (
            [],
            [0, 80],
            [],
            [63],
            [],
            [8, 55],
            [],
            [42],
            [],
        ):
            store1.DelRanks(deletion)
            for rank in deletion:
                del strata[ranks.index(rank)]
                ranks.remove(rank)
            assert [
                *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
            ] == [*store1.IterRankDifferentiaZip()]

    def test_DelRanks_getrank_impl4(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 55, 63]
        strata = [
            hstrat.HereditaryStratum(deposition_rank=rank) for rank in ranks
        ]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        for deletion in (
            [],
            [55, 63],
            [],
            [0, 8],
            [],
            [42],
            [],
        ):
            store1.DelRanks(deletion)
            for rank in deletion:
                del strata[ranks.index(rank)]
                ranks.remove(rank)
            assert [
                *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
            ] == [*store1.IterRankDifferentiaZip()]

    def test_DelRanks_getrank_impl5(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 55, 63]
        strata = [
            hstrat.HereditaryStratum(deposition_rank=rank) for rank in ranks
        ]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        for deletion in (
            [],
            [42, 63],
            [],
            [0, 8, 55],
            [],
        ):
            store1.DelRanks(deletion)
            for rank in deletion:
                del strata[ranks.index(rank)]
                ranks.remove(rank)
            assert [
                *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
            ] == [*store1.IterRankDifferentiaZip()]

    def test_DelRanks_calcrank_impl1(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 55, 63]
        strata = [hstrat.HereditaryStratum() for rank in ranks]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        for deletion in (
            [],
            [8, 42],
            [],
            [55],
            [],
            [0, 63],
            [],
        ):
            store1.DelRanks(
                get_column_index_of_rank=ranks.index,
                ranks=deletion,
            )
            for rank in deletion:
                del strata[ranks.index(rank)]
                ranks.remove(rank)
            assert [
                *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
            ] == [
                *store1.IterRankDifferentiaZip(
                    get_rank_at_column_index=lambda idx: ranks[idx],
                )
            ]

    def test_DelRanks_calcrank_impl2(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 55, 63]
        strata = [hstrat.HereditaryStratum() for rank in ranks]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        for deletion in (
            [],
            [0, 63],
            [],
            [8],
            [],
            [55],
            [],
            [42],
            [],
        ):
            store1.DelRanks(
                get_column_index_of_rank=ranks.index,
                ranks=deletion,
            )
            for rank in deletion:
                del strata[ranks.index(rank)]
                ranks.remove(rank)
            assert [
                *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
            ] == [
                *store1.IterRankDifferentiaZip(
                    get_rank_at_column_index=lambda idx: ranks[idx],
                )
            ]

    def test_DelRanks_calcrank_impl3(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 55, 63, 80]
        strata = [hstrat.HereditaryStratum() for rank in ranks]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        for deletion in (
            [],
            [0, 80],
            [],
            [63],
            [],
            [8, 55],
            [],
            [42],
            [],
        ):
            store1.DelRanks(
                get_column_index_of_rank=ranks.index,
                ranks=deletion,
            )
            for rank in deletion:
                del strata[ranks.index(rank)]
                ranks.remove(rank)
            assert [
                *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
            ] == [
                *store1.IterRankDifferentiaZip(
                    get_rank_at_column_index=lambda idx: ranks[idx],
                )
            ]

    def test_DelRanks_calcrank_impl4(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 55, 63]
        strata = [hstrat.HereditaryStratum() for rank in ranks]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        for deletion in (
            [],
            [55, 63],
            [],
            [0, 8],
            [],
            [42],
            [],
        ):
            store1.DelRanks(
                get_column_index_of_rank=ranks.index,
                ranks=deletion,
            )
            for rank in deletion:
                del strata[ranks.index(rank)]
                ranks.remove(rank)
            assert [
                *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
            ] == [
                *store1.IterRankDifferentiaZip(
                    get_rank_at_column_index=lambda idx: ranks[idx],
                )
            ]

    def test_DelRanks_calcrank_impl5(self):
        store1 = hstrat.HereditaryStratumOrderedStoreList()
        ranks = [0, 8, 42, 55, 63]
        strata = [hstrat.HereditaryStratum() for rank in ranks]
        for rank, stratum in zip(ranks, strata):
            store1.DepositStratum(rank=rank, stratum=stratum)

        for deletion in (
            [],
            [42, 63],
            [],
            [0, 8, 55],
            [],
        ):
            store1.DelRanks(
                get_column_index_of_rank=ranks.index,
                ranks=deletion,
            )
            for rank in deletion:
                del strata[ranks.index(rank)]
                ranks.remove(rank)
            assert [
                *zip(ranks, [stratum.GetDifferentia() for stratum in strata])
            ] == [
                *store1.IterRankDifferentiaZip(
                    get_rank_at_column_index=lambda idx: ranks[idx],
                )
            ]


if __name__ == "__main__":
    unittest.main()



import argparse
import logging
import warnings

from joinem._dataframe_cli import _add_parser_base, _run_dataframe_cli
import pandas as pd
import polars as pl

from ._alifestd_make_ancestor_list_col import alifestd_make_ancestor_list_col
from ._alifestd_try_add_ancestor_list_col_polars import (
    alifestd_try_add_ancestor_list_col_polars,
)
from ._configure_prod_logging import configure_prod_logging
from ._delegate_polars_implementation import (
    DataFrame_T,
    delegate_polars_implementation,
)
from ._format_cli_description import format_cli_description
from ._get_hstrat_version import get_hstrat_version
from ._log_context_duration import log_context_duration


@delegate_polars_implementation(alifestd_try_add_ancestor_list_col_polars)
def alifestd_try_add_ancestor_list_col(
    phylogeny_df: DataFrame_T,
    root_ancestor_token: str = "none",
    mutate: bool = False,
) -> DataFrame_T:
    """Add an ancestor_list column to the input DataFrame if the column does
    not already exist.

    Input dataframe is not mutated by this operation unless `mutate` set True.
    If mutate set True, operation does not occur in place; still use return
    value to get transformed phylogeny dataframe.

    See Also
    --------
    alifestd_make_ancestor_list_col
    """

    assert isinstance(phylogeny_df, pd.DataFrame)
    if "ancestor_id" in phylogeny_df and "ancestor_list" not in phylogeny_df:
        logging.info("ancestor_id column present, adding ancestor_list column")
        if not mutate:
            phylogeny_df = phylogeny_df.copy()
        phylogeny_df["ancestor_list"] = alifestd_make_ancestor_list_col(
            phylogeny_df["id"],
            phylogeny_df["ancestor_id"],
            root_ancestor_token=root_ancestor_token,
        )
    elif "ancestor_list" in phylogeny_df:
        logging.info("ancestor_list column already present, skipping addition")
    else:
        logging.info(
            "no ancestor_id column available, skipping ancestor_list addition",
        )

    return phylogeny_df


_raw_description = """Create 'ancestor_list' column, if not already present, to comply with Alife standard phylogeny data format.

Additional Notes
================
- Requires 'ancestor_id' column to be present in input DataFrame.
Otherwise, no action is taken.

- Use `--eager-read` if modifying data file inplace.

- This CLI entrypoint is experimental and may be subject to change.
"""


def _create_parser() -> argparse.ArgumentParser:
    """Create parser for CLI entrypoint."""
    parser = argparse.ArgumentParser(
        description=format_cli_description(_raw_description),
        formatter_class=argparse.RawTextHelpFormatter,
    )
    _add_parser_base(
        parser=parser,
        dfcli_module="hstrat._auxiliary_lib._alifestd_try_add_ancestor_list_col",
        dfcli_version=get_hstrat_version(),
    )
    return parser


def _checked_alifestd_add_ancestor_list_col_polars(
    df: pl.DataFrame,
) -> pl.DataFrame:
    """Wrapper for `alifestd_try_add_ancestor_list_col` that checks for
    required columns, preventing silent failure."""

    if "ancestor_id" not in df and "ancestor_list" not in df:
        warnings.warn(
            "Creating 'ancestor_list' column requires 'ancestor_id' column, "
            "but it is not provided.",
        )

    return alifestd_try_add_ancestor_list_col_polars(df)


if __name__ == "__main__":
    configure_prod_logging()

    parser = _create_parser()
    args, __ = parser.parse_known_args()

    with log_context_duration(
        "hstrat._auxiliary_lib._alifestd_try_add_ancestor_list_col",
        logging.info,
    ):
        _run_dataframe_cli(
            base_parser=parser,
            output_dataframe_op=_checked_alifestd_add_ancestor_list_col_polars,
        )



import typing

import opytional as opyt
import pandas as pd

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._impl import build_tree_searchtable_python

try:
    from ._impl._build_tree_searchtable_cpp import build_tree_searchtable_cpp
except ImportError:  # pragma: no cover
    build_tree_searchtable_cpp = None


def build_tree_searchtable(
    population: typing.Sequence[HereditaryStratigraphicArtifact],
    taxon_labels: typing.Optional[typing.Iterable] = None,
    progress_wrap: typing.Optional[typing.Callable] = None,
    force_common_ancestry: bool = False,
    use_impl: typing.Literal["cpp", "python", None] = None,
) -> pd.DataFrame:
    """Estimate the phylogenetic history among hereditary stratigraphic
    artifacts by building a trie (a.k.a. prefix tree) of their differentiae
    records.

    In contrast to `build_tree_trie`, this function applies a more efficient
    algorithm that collapses away inner nodes corresponding to dropped ranks
    to prevent exponential time complexity from "wildcard" (i.e., missing)
    differentiae when reconstructing the phylogeny.

    Parameters
    ----------
    population: Sequence[HereditaryStratigraphicArtifact]
        Hereditary stratigraphic columns corresponding to extant population members.

        Each member of population will correspond to a unique leaf node in the
        reconstructed tree.
    taxon_labels: Optional[Iterable], optional
        How should leaf nodes representing extant hereditary stratigraphic
        columns be named?

        Label order should correspond to the order of corresponding hereditary
        stratigraphic columns within `population`. If None, taxons will be
        named according to their numerical index.
    force_common_ancestry: bool, default False
        How should columns that definitively share no common ancestry be
        handled?

        If set to True, treat columns with no common ancestry as if they
        shared a common ancestor immediately before the genesis of the
        lineages. If set to False, columns within `population` that
        definitively do not share common ancestry will raise a ValueError.
    progress_wrap : Callable, optional
        Pass tqdm or equivalent to display a progress bar.
    use_impl : {'cpp', 'python', None}, optional
        Which implementation to use.

        If None, the best available implementation will be used.
        If 'cpp', the C++ implementation will be used.
        If 'python', the Python implementation will be used.

    Returns
    -------
    pd.DataFrame
        The reconstructed phylogenetic tree in alife standard format.

    Notes
    -----
    Unifurcations in the reconstructed tree are collapsed.

    However, polytomies are not resolved. In addition to any true polytomies,
    ancestry sequences that cannot be resolved due to missing information
    appear as polytomies in the generated reconstruction. Therefore, polytomies
    are generally overrepresented in reconstructions, especially when low
    hereditary stratigraphic resolution is available. If overestimation of
    polytomies is problematic, external tools can be used to decompose
    polytomies into arbitrarily-arranged bifurcations.

    Uses a more efficient tabular, rather than node-and-pointer, underlying trie
    representation. Two trie structures are maintained: one for building the
    trie, and one for searching the trie. The "search" trie structure is
    reconfigurable to collapse portions of the trie for efficient search. The
    "build" trie structure is used to record the original structure of the trie.

    See Also
    --------
    build_tree_trie : Naive trie-based phylogenetic reconstruction algorithm.
    """
    try:
        build_tree_searchtable_impl = {
            "cpp": build_tree_searchtable_cpp,
            "either": opyt.or_value(
                build_tree_searchtable_cpp,
                build_tree_searchtable_python,
            ),
            "python": build_tree_searchtable_python,
        }[opyt.or_value(use_impl, "either")]
    except KeyError:
        raise ValueError(
            f"Invalid value {use_impl} for `use_impl`. "
            "Expected one of 'cpp', 'python', or None.",
        )

    if build_tree_searchtable_impl is None:  # pragma: no cover
        raise ImportError(
            f"build_tree_searchtable impl '{use_impl}' is unavailable.",
        )

    return build_tree_searchtable_impl(
        population,
        taxon_labels,
        progress_wrap,
        force_common_ancestry,
    )



import logging
import math

from downstream import dataframe as dstream_dataframe
import pandas as pd
import polars as pl
import tqdm

from .._auxiliary_lib import (
    alifestd_make_empty,
    get_sole_scalar_value_polars,
    log_context_duration,
    log_memory_usage,
    render_polars_snapshot,
)
from ..phylogenetic_inference.tree._impl._build_tree_searchtable_cpp_impl_stub import (
    Records,
    collapse_dropped_unifurcations,
    extend_tree_searchtable_cpp_from_exploded,
    records_to_dict,
)


def _build_records_chunked(
    df: pl.DataFrame, exploded_slice_size: int
) -> Records:
    """Build tree searchtable from DataFrame, exploding in chunks to reduce
    memory usage."""
    num_slices = math.ceil(len(df) / exploded_slice_size)
    logging.info(f"{len(df)=} {exploded_slice_size=} {num_slices=}")

    init_size = len(df) * 8
    records = Records(init_size)
    for i, df_slice in enumerate(df.iter_slices(exploded_slice_size)):
        logging.info(f"incorporating slice {i + 1}/{num_slices}...")

        with log_context_duration(
            f"dstream.dataframe.explode_lookup_unpacked ({i + 1}/{num_slices})",
            logging.info,
        ):
            long_df = dstream_dataframe.explode_lookup_unpacked(
                df_slice, value_type="uint64"
            ).select(
                pl.col(
                    "dstream_data_id",
                    "dstream_T",
                    "dstream_Tbar",
                    "dstream_value",
                )
                .sort_by("dstream_Tbar")
                .over(partition_by="dstream_data_id"),
            )

        if i == 0:
            render_polars_snapshot(long_df, "exploded", logging.info)

        with log_context_duration(
            f"extend_tree_searchtable_cpp_from_exploded ({i + 1}/{num_slices})",
            logging.info,
        ):
            extend_tree_searchtable_cpp_from_exploded(
                records,
                long_df["dstream_data_id"].to_numpy(),
                long_df["dstream_T"].to_numpy(),
                long_df["dstream_Tbar"].to_numpy(),
                long_df["dstream_value"].to_numpy(),
                tqdm.tqdm,
            )

        with log_context_duration(
            f"collapse_dropped_unifurcations ({i + 1}/{num_slices})",
            logging.info,
        ):
            records = collapse_dropped_unifurcations(records)

        log_memory_usage(logging.info)

    return records


def _join_user_defined_columns(
    df: pl.DataFrame, phylo_df: pl.DataFrame
) -> pl.DataFrame:
    """Join user-defined columns from input data onto reconstructed tree
    dataframe."""
    df = df.select(
        pl.exclude("^dstream_.*$", "^downstream_.*$"),
        pl.col("dstream_data_id").cast(pl.UInt64),
    )
    joined_columns = set(df.columns) - set(phylo_df.columns)
    if joined_columns:
        logging.info(f" - {len(joined_columns)} column to join")
        logging.info(f" - joining columns: {[*joined_columns]}")
        phylo_df = phylo_df.join(df, on="dstream_data_id", how="left")
    else:
        logging.info(" - no columns to join, skipping")

    return phylo_df


def _construct_result_dataframe(
    records: Records,
    differentia_bitwidth: int,
    dstream_S: int,
) -> pl.DataFrame:
    """Convert tree searchtable records to DataFrame."""
    logging.info("converting records to dict...")
    records_dict = records_to_dict(records)

    logging.info("converting dict to dataframe...")
    del records_dict["differentia"]
    return (
        pl.from_dict(
            records_dict,  # type: ignore
            schema={
                "dstream_data_id": pl.UInt64,
                "id": pl.UInt64,
                "ancestor_id": pl.UInt64,
                "rank": pl.UInt64,
            },
        )
        .with_columns(
            pl.lit(differentia_bitwidth)
            .alias("hstrat_differentia_bitwidth")
            .cast(pl.UInt32),
            pl.lit(dstream_S).alias("dstream_S").cast(pl.UInt32),
        )
        .rename({"rank": "hstrat_rank"})
    )


def _surface_unpacked_reconstruct(
    df: pl.DataFrame,
    *,
    differentia_bitwidth: int,
    dstream_S: int,
    exploded_slice_size: int,
) -> pl.DataFrame:
    """Reconstruct phylogenetic tree from unpacked dstream data."""
    render_polars_snapshot(df, "unpacked", logging.info)

    logging.info("building tree searchtable chunkwise...")
    records = _build_records_chunked(df, exploded_slice_size)

    with log_context_duration("_construct_result_dataframe", logging.info):
        phylo_df = _construct_result_dataframe(
            records,
            differentia_bitwidth=differentia_bitwidth,
            dstream_S=dstream_S,
        )

    del records
    render_polars_snapshot(phylo_df, "converted", logging.info)

    logging.info("joining user-defined columns...")
    with log_context_duration("_join_user_defined_columns", logging.info):
        phylo_df = _join_user_defined_columns(df, phylo_df)

    logging.info("surface_unpack_reconstruct complete")
    render_polars_snapshot(phylo_df, "reconstruction", logging.info)

    return phylo_df


def surface_unpack_reconstruct(
    df: pl.DataFrame,
    *,
    exploded_slice_size: int = 1_000_000,
) -> pl.DataFrame:
    """Unpack dstream buffer and counter from genome data and construct an
    estimated phylogenetic tree for the genomes.

    Parameters
    ----------
    df : pl.DataFrame
        The input DataFrame containing packed data with required columns, one
        row per genome.

        Required schema:
            - 'data_hex' : pl.String
                - Raw genome data, with serialized dstream buffer and counter.
                - Represented as a hexadecimal string.
            - 'dstream_algo' : pl.Categorical
                - Name of downstream curation algorithm used.
                - e.g., 'dstream.steady_algo'
            - 'dstream_storage_bitoffset' : pl.UInt64
                - Position of dstream buffer field in 'data_hex'.
            - 'dstream_storage_bitwidth' : pl.UInt64
                - Size of dstream buffer field in 'data_hex'.
            - 'dstream_T_bitoffset' : pl.UInt64
                - Position of dstream counter field in 'data_hex'.
            - 'dstream_T_bitwidth' : pl.UInt64
                - Size of dstream counter field in 'data_hex'.
            - 'dstream_S' : pl.Uint32
                - Capacity of dstream buffer, in number of data items.

        Optional schema:
            - 'downstream_version' : pl.Categorical
                - Version of downstream library used to curate data items.
            - 'dstream_data_id' : pl.UInt64
                - Unique identifier for each data item.
                - If not provided, row number will be used as identifier.
            - 'downstream_exclude_exploded' : pl.Boolean
                - Should row be dropped after exploding unpacked data?
            - 'downstream_exclude_unpacked' : pl.Boolean
                - Should row be dropped after unpacking packed data?
            - 'downstream_validate_exploded' : pl.String, polars expression
                - Polars expression to validate exploded data.
            - 'downstream_validate_unpacked' : pl.String, polars expression
                - Polars expression to validate unpacked data.

    exploded_slice_size : int, default 1_000_000
        Number of rows to process at once. Lower values reduce memory usage.

    Returns
    -------
    pl.DataFrame
        The output DataFrame containing the estimated phylogenetic tree in
        alife standard format, with the following columns:

        - 'id' : pl.UInt64
            - Unique identifier for each taxon (RE alife standard format).
        - 'ancestor_id' : pl.UInt64
            - Unique identifier for ancestor taxon  (RE alife standard format).
        - 'hstrat_rank' : pl.UInt64
            - Num generations elapsed for ancestral differentia.
            - Corresponds to`dstream_Tbar` for inner nodes.
            - Corresponds `dstream_T` - 1 for leaf nodes
        - 'hstrat_differentia_bitwidth' : pl.UInt64
            - Size of annotation differentiae, in bits.
            - Corresponds to `dstream_value_bitwidth`.
        - 'dstream_S' : pl.UInt32
            - Capacity of dstream buffer, in number of data items.
        - 'dstream_data_id' : pl.UInt64
            - Unique identifier for each genome in source dataframe
            - Set to source dataframe row index if not provided.

        User-defined columns, except some prefixed with 'downstream_' or
        'dstream_', will be forwarded from the input DataFrame.

        Note that the alife-standard `ancestor_list` column is not included in
        the output.

    See Also
    --------
    surface_postprocess_trie :
        Post-processes raw phylogeny data, including collapsing superfluous
        internal nodes and estimating taxon origin times.
    alifestd_try_add_ancestor_list_col :
        Adds alife-standard `ancestor_list` column to phylogeny data.
    """
    logging.info("beginning surface_unpack_reconstruct")
    log_memory_usage(logging.info)

    render_polars_snapshot(df, "packed", logging.info)

    # for simplicity, return early for this special case
    if df.lazy().limit(1).collect().is_empty():
        logging.info("empty input dataframe, returning empty result")
        res = alifestd_make_empty()
        res["taxon_label"] = None
        res["hstrat_rank"] = pd.Series(dtype=int)
        res["hstrat_differentia_bitwidth"] = pd.Series(dtype=int)
        res["dstream_S"] = pd.Series(dtype=int)
        return pl.from_pandas(res)

    logging.info("extracting metadata..")
    dstream_storage_bitwidth = get_sole_scalar_value_polars(
        df, "dstream_storage_bitwidth"
    )
    logging.info(f" - dstream storage bitwidth: {dstream_storage_bitwidth}")

    dstream_S = get_sole_scalar_value_polars(df, "dstream_S")
    logging.info(f" - dstream S: {dstream_S}")

    if not dstream_storage_bitwidth % dstream_S == 0:
        raise ValueError(
            "dstream_storage_bitwidth must be a multiple of dstream_S, "
            "cannot calculate differentia_bitwidth",
        )
    differentia_bitwidth = dstream_storage_bitwidth // dstream_S
    logging.info(f" - differentia bitwidth: {differentia_bitwidth}")

    with log_context_duration(
        "dstream.dataframe.unpack_data_packed", logging.info
    ):
        df = dstream_dataframe.unpack_data_packed(df)

    return _surface_unpacked_reconstruct(
        df,
        differentia_bitwidth=differentia_bitwidth,
        dstream_S=dstream_S,
        exploded_slice_size=exploded_slice_size,
    )



import operator
import typing

from ..._auxiliary_lib import HereditaryStratigraphicArtifact
from ._calc_rank_of_earliest_detectable_mrca_between import (
    calc_rank_of_earliest_detectable_mrca_between,
)
from ._calc_ranks_since_mrca_bounds_with import (
    calc_ranks_since_mrca_bounds_with,
)


def calc_ranks_since_mrca_uncertainty_with(
    focal: HereditaryStratigraphicArtifact,
    other: HereditaryStratigraphicArtifact,
    prior: typing.Literal["arbitrary"],
    confidence_level: float = 0.95,
) -> typing.Optional[int]:
    """How wide is the estimation window for generations elapsed since MRCA?

    Calculates uncertainty of estimate for the number of depositions
    elapsed along focal column's line of descent since the most common recent
    ancestor with other.

    Returns 0 if no common ancestor between focal and other can be resolved
    with sufficient confidence. If insufficient common ranks between focal
    and other are available to resolve any common ancestor, returns None.

    See Also
    --------
    calc_ranks_since_mrca_bounds_with :
        Calculates bound whose uncertainty this method reports. See the
        corresponding docstring for explanation of parameters.
    """
    assert 0.0 <= confidence_level <= 1.0

    if (
        calc_rank_of_earliest_detectable_mrca_between(
            focal,
            other,
            confidence_level=confidence_level,
        )
        is None
    ):
        return None

    bounds = calc_ranks_since_mrca_bounds_with(
        focal,
        other,
        prior=prior,
        confidence_level=confidence_level,
    )
    return 0 if bounds is None else abs(operator.sub(*bounds)) - 1



import polars as pl

from hstrat._auxiliary_lib import render_polars_snapshot


def test_smoke():
    df = pl.DataFrame(
        {
            "foo": [1, 2, 3],
            "bar": [6, 7, 8],
            "ham": ["buzz", "baz", "bang"],
        }
    )
    assert "buzz" in render_polars_snapshot(df)
    assert "bazinga" in render_polars_snapshot(df, "bazinga")
    assert "zzub" in render_polars_snapshot(df, display=lambda x: x[::-1])
